#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np

def calc_part_src_nbi(beam, iol_adjusted=False, F_orb_nbi=None):
    """
    """
    # # old suspicious calculation
    # snbi = (0.624E25 * beam.dPdV.v1D.W / (vol * beam.E.J)) * beam.num * beam.dp

    # Piper Changes: Changed function to calculate particle source and source density.
    # tot, lost, and kept are all densities. The first return variable is a total source.
    # particle source needs to be split into a total, amount kept, and amount lost to calculate the return current.

    # These equations assume:
    # beam.dPdV.v1D.W = H(rho) * Pbeam * pwrfrac / Volp_nbi
    # beam.dPdr.v1D.W = beam.dPdV.v1D.W * dV

    # The factor of 2 can be applied to F_orb in the continuity equation later.

    # nbi source in particles/sec. Needed to calculate gamma with sources.
    try:

        part_src_nbi = sum([beam.dPdr.v1D.W[i] / beam.E.J[i] for i in range(len(beam.E.J))])

        # nbi source in # of particles/(m^3 * sec). Needed to calculate gamma with source densities.

        part_src_nbi_tot = sum([beam.dPdV.v1D.W[i] / beam.E.J[i] for i in range(len(beam.E.J))])

    except:
        part_src_nbi = beam.dPdr.v1D.W / beam.E.J
        part_src_nbi_tot = beam.dPdV.v1D.W / beam.E.J


    if iol_adjusted:
        part_src_nbi = part_src_nbi * (1 - F_orb_nbi)
        part_src_nbi_lost = part_src_nbi_tot * (F_orb_nbi)
        part_src_nbi_kept = part_src_nbi_tot * (1 - F_orb_nbi)
    else:
        part_src_nbi_lost = np.zeros(part_src_nbi_tot.shape)
        part_src_nbi_kept = part_src_nbi_tot

    return part_src_nbi, part_src_nbi_tot, part_src_nbi_lost, part_src_nbi_kept



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from scipy.constants import constants

e = constants.elementary_charge
z_c = 6  # atomic number of carbon

def calc_vpol(Er, vphi_j, Lp, T, n, z_d, B_t, B_p, vphi_k, vpol_k, z_k):
    vpol = (1.0/B_t) * (1.0/(e*z_d) * -Lp.i * T.i.J + vphi_j * B_p - Er)
    vpol_assum = vpol_k - (T.C.J / (e * B_t)) * (Lp.i - (Lp.C / z_c)) + (B_p / B_t) * (vphi_j - vphi_k)
    vpol_alt = vpol_k - 1.0 / (e * B_t) * (T.i.J * Lp.i - (T.C.J * Lp.C / z_c)) + (B_p / B_t) * (vphi_j - vphi_k)
    return vpol, vpol_assum, vpol_alt



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri May 18 15:14:01 2018

@author: max
"""
from __future__ import division
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import griddata, UnivariateSpline
from scipy.constants import elementary_charge
from shapely.geometry import Point, LineString
from shapely.ops import polygonize
from collections import namedtuple
from math import ceil
from GT3.Core.Functions.Cut import cut
from contours.quad import QuadContourGenerator
import sys


class Sol:
    def __init__(self, inp, core):
        # # NOTE: this assumes structured psi data on a regular grid. This way of getting R and Z is entirely inadeqaute for
        # # unstructured psi data
        # R = inp.psirz_exp[:, 0].reshape(-1, np.unique(inp.psirz_exp[:, 0]).size)
        # Z = inp.psirz_exp[:, 1].reshape(-1, np.unique(inp.psirz_exp[:, 1]).size)

        self.calc_sol_lines(inp, core)
        self.calc_sol_nT(inp, core)
        pass

    def calc_sol_lines(self, inp, core):
        c = QuadContourGenerator.from_rectilinear(core.psi_data.R[0], core.psi_data.Z[:, 0], core.psi_data.psi_norm)
        self.sol_lines = []
        self.sol_lines_cut = []

        #
        # first we need to make sure that the inp.sollines_psi_max specified in the input file doesn't
        # go outside of the first wall. For now, we will check for this as follows:
        #   1. draw the contours for psi_norm = inp.sollines_psi_max
        #   2. if there is only 1, make sure it only intersects the first wall no more than twice
        #   3. if there is only 1 and it intersects the first wall more than twice, raise an error
        #       that tells the user to reduce their sollines_psi_max value in the input file
        #   4. if there is more than 1 contour, then we need to determine which one is the correct one.
        #       for now, we will check that it's largest y value is higher than the magnetic axis,
        #       it's largest x value is to the right of the magnetic axis, it's lowest y value is lower
        #       than the magnetic axis, and it's lowest x value is to the left of the magnetic axis.
        #   5. Once the correct contour has been identified, do steps 2 and 3.
        #

        sollines_psi_max_contours = c.contour(inp.sollines_psi_max)
        num_lines = len(sollines_psi_max_contours)

        print 'num_lines = ', num_lines
        if num_lines == 1:
            # then this is probably the correct line. Check to see how many times it intersects
            # with the first wall
            num_wall_ints = len(LineString(sollines_psi_max_contours[0]).intersection(inp.wall_line))
            print 'num_wall_ints = ', num_wall_ints
            if num_wall_ints > 2:
                print 'It looks like your sollines_psi_max value might be intersecting the wall.' \
                      'Try reducing it. Stopping.'
                raise Exception("It looks like your sollines_psi_max value might be intersecting the wall. Try reducing it.")
        else:
            for i, line in enumerate(sollines_psi_max_contours):
                max_x = np.amax(line[:, 0])
                min_x = np.amin(line[:, 0])
                max_y = np.amax(line[:, 1])
                min_y = np.amin(line[:, 1])
                if min_x < core.pts.axis.mag[0] < max_x and min_y < core.pts.axis.mag[1] < max_y:
                    # then this is probably the correct line. Check to see how many times it intersects
                    # with the first wall
                    num_wall_ints = len(LineString(line).intersection(inp.wall_line))
                    if num_wall_ints > 2:
                        print 'It looks like your sollines_psi_max value might be intersecting the wall.' \
                              'Try reducing it. Stopping.'
                        raise Exception("It looks like your sollines_psi_max value might be intersecting the wall. Try reducing it.")
                    else:
                        break

        psi_pts = np.linspace(1, inp.sollines_psi_max, inp.num_sollines + 1, endpoint=True)[1:]

        for i, v in enumerate(psi_pts):
            num_lines = len(c.contour(v))
            if num_lines == 1:
                # this is probably the correct line
                self.sol_lines.append(LineString(c.contour(v)[0]))
            else:
                # we need to determine which contour to use
                for line in c.contour(v):
                    max_x = np.amax(line[:, 0])
                    min_x = np.amin(line[:, 0])
                    max_y = np.amax(line[:, 1])
                    min_y = np.amin(line[:, 1])
                    if min_x < core.pts.axis.mag[0] < max_x and min_y < core.pts.axis.mag[1] < max_y:
                        # then this is probably the correct line.
                        self.sol_lines.append(LineString(line))

        for line in self.sol_lines:
            # find intersection points with the wall
            int_pts = line.intersection(inp.wall_line)
            # cut line at intersection points
            cut_line = cut(line, line.project(int_pts[0], normalized=True))[1]
            cut_line = cut(cut_line, cut_line.project(int_pts[1], normalized=True))[0]
            self.sol_lines_cut.append(cut_line)

        # add wall intersection points from divertor legs and sol lines to wall_line.
        # This is necessary to prevent thousands of tiny triangles from forming if the
        # end of the flux line isn't exactly on top of the wall line.

        # add inboard seperatrix strike point

        # plt.plot(np.asarray(inp.wall_line)[:,0], np.asarray(inp.wall_line)[:,1])
        # plt.plot(np.asarray(core.lines.div.ib_long)[:,0], np.asarray(core.lines.div.ib_long)[:, 1])

        union = inp.wall_line.union(core.lines.div.ib_long)
        result = [geom for geom in polygonize(union)][0]
        inp.wall_line = LineString(result.exterior.coords)

        # add outboard seperatrix strike point
        union = inp.wall_line.union(core.lines.div.ob_long)
        result = [geom for geom in polygonize(union)][0]
        inp.wall_line = LineString(result.exterior.coords)

        # add sol line intersection points on inboard side
        # for some reason, union freaks out when I try to do inboard and outboard
        # at the same time.
        for num, line in enumerate(self.sol_lines):
            union = inp.wall_line.union(cut(line, 0.5)[0])
            result = [geom for geom in polygonize(union)][0]
            inp.wall_line = LineString(result.exterior.coords)

        # add sol line intersection points on outboard side
        for num, line in enumerate(self.sol_lines):
            union = inp.wall_line.union(cut(line, 0.5)[1])
            result = [geom for geom in polygonize(union)][0]
            inp.wall_line = LineString(result.exterior.coords)

    def calc_sol_nT(self, inp, core):
        # calculate spatial gradients for density and temperature along the seperatrix from dni/dr = dni/dpsi * dpsi/dr
        # specify the flux surface to get densities, temperatures, and their gradients
        sep_flx_surf = .98

        # calculate dni/dpsi and dTi/dpsi at the seperatrix
        # TODO: include both deuterium and tritium here
        ni_psi_fit = UnivariateSpline(core.rho2psi(core.rho[:, 0]), core.n.i[:, 0], k=3, s=2.0)
        ne_psi_fit = UnivariateSpline(core.rho2psi(core.rho[:, 0]), core.n.e[:, 0], k=3, s=2.0)
        Ti_psi_fit = UnivariateSpline(core.rho2psi(core.rho[:, 0]), core.T.i.kev[:, 0], k=3, s=2.0)
        Te_psi_fit = UnivariateSpline(core.rho2psi(core.rho[:, 0]), core.T.e.kev[:, 0], k=3, s=2.0)
        dni_dpsi_sep = ni_psi_fit.derivative()(sep_flx_surf)
        dTi_dpsi_sep = Ti_psi_fit.derivative()(sep_flx_surf)

        # calculate dpsidr everywhere (technically, we're calculating |dpsi/dr|. We don't care about the direction.
        dpsidR = np.abs(np.gradient(core.psi_data.psi_norm, core.psi_data.R[0, :], axis=1))
        dpsidZ = np.abs(np.gradient(core.psi_data.psi_norm, core.psi_data.Z[:, 0], axis=0))

        dpsidr = dpsidR + dpsidZ

        dpsidr_sep = griddata(np.column_stack((core.psi_data.R.flatten(), core.psi_data.Z.flatten())),
                              dpsidr.flatten(),
                              np.asarray(core.lines.sep.coords),
                              method='linear')

        # calculate dni/dr and dTi/dr at the seperatrix
        dnidr_sep = dni_dpsi_sep * dpsidr_sep
        dnedr_sep = dnidr_sep
        dTidr_sep = dTi_dpsi_sep * dpsidr_sep
        dTedr_sep = dTidr_sep

        num_sep_pts = len(np.asarray(core.lines.sep.coords))

        ni_sep = np.full(num_sep_pts, ni_psi_fit(sep_flx_surf))
        ne_sep = np.full(num_sep_pts, ne_psi_fit(sep_flx_surf))
        Ti_sep = np.full(num_sep_pts, Ti_psi_fit(sep_flx_surf)) * 1E3 * 1.6021E-19  # in Joules
        Te_sep = np.full(num_sep_pts, Te_psi_fit(sep_flx_surf)) * 1E3 * 1.6021E-19  # in Joules

        # calculate BT along the seperatrix
        BT_sep = inp.BT0 * core.pts.axis.mag[0] / np.asarray(core.lines.sep.coords)[:, 0]

        # remove a certain percentage of the seperatrix in the vicinity of the x-point
        # this fraction will be removed from both the inboard and the outboard sides of the x-point

        frac_to_remove = 0.1
        pts_to_remove = int(ceil(frac_to_remove * num_sep_pts))

        dnidr_sep_cut = dnidr_sep[pts_to_remove:-pts_to_remove]
        dnedr_sep_cut = dnedr_sep[pts_to_remove:-pts_to_remove]
        dTidr_sep_cut = dTidr_sep[pts_to_remove:-pts_to_remove]
        dTedr_sep_cut = dTedr_sep[pts_to_remove:-pts_to_remove]

        ni_sep_cut = ni_sep[pts_to_remove:-pts_to_remove]
        ne_sep_cut = ne_sep[pts_to_remove:-pts_to_remove]
        Ti_sep_cut = Ti_sep[pts_to_remove:-pts_to_remove]
        Te_sep_cut = Te_sep[pts_to_remove:-pts_to_remove]

        BT_sep_cut = BT_sep[pts_to_remove:-pts_to_remove]

        # define densities and temperatures along divertor legs
        # TODO: Specify these things in the input file
        ni_ib_wall = ni_sep_cut[0] * 10
        ni_ob_wall = ni_sep_cut[-1] * 10
        ni_ib = np.linspace(ni_ib_wall, ni_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        ni_ob = np.linspace(ni_sep_cut[-1], ni_ob_wall, inp.xi_ob_pts, endpoint=True)

        ne_ib_wall = ne_sep_cut[0] * 10
        ne_ob_wall = ne_sep_cut[-1] * 10
        ne_ib = np.linspace(ne_ib_wall, ne_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        ne_ob = np.linspace(ne_sep_cut[-1], ne_ob_wall, inp.xi_ob_pts, endpoint=True)

        Ti_ib_wall = Ti_sep_cut[0] / 4
        Ti_ob_wall = Ti_sep_cut[-1] / 4
        Ti_ib = np.linspace(Ti_ib_wall, Ti_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        Ti_ob = np.linspace(Ti_sep_cut[-1], Ti_ob_wall, inp.xi_ob_pts, endpoint=True)

        Te_ib_wall = Te_sep_cut[0] / 4
        Te_ob_wall = Te_sep_cut[-1] / 4
        Te_ib = np.linspace(Te_ib_wall, Te_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        Te_ob = np.linspace(Te_sep_cut[-1], Te_ob_wall, inp.xi_ob_pts, endpoint=True)

        print
        print '#####################################'
        print ' divertor values'
        print ' ni_ib_wall = ', ni_ib_wall
        print ' ni_ob_wall = ', ni_ob_wall
        print ' ne_ib_wall = ', ne_ib_wall
        print ' ne_ob_wall = ', ne_ob_wall
        print ' Ti_ib_wall(ev) = ', Ti_ib_wall / 1.6021E-19
        print ' Ti_ob_wall(ev) = ', Ti_ob_wall / 1.6021E-19
        print ' Te_ib_wall(ev) = ', Te_ib_wall / 1.6021E-19
        print ' Te_ob_wall(ev) = ', Te_ob_wall / 1.6021E-19
        print '#####################################'
        print

        # define density and temperature gradients along the inboard and outboard divertor legs
        dnidr_ib_wall = dnidr_sep_cut[0]
        dnidr_ob_wall = dnidr_sep_cut[-1]
        dnidr_ib = np.linspace(dnidr_ib_wall, dnidr_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        dnidr_ob = np.linspace(dnidr_sep_cut[-1], dnidr_ob_wall, inp.xi_ob_pts, endpoint=True)

        dnedr_ib_wall = dnedr_sep_cut[0]
        dnedr_ob_wall = dnedr_sep_cut[-1]
        dnedr_ib = np.linspace(dnedr_ib_wall, dnedr_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        dnedr_ob = np.linspace(dnedr_sep_cut[-1], dnedr_ob_wall, inp.xi_ob_pts, endpoint=True)

        dTidr_ib_wall = dTidr_sep_cut[0]
        dTidr_ob_wall = dTidr_sep_cut[-1]
        dTidr_ib = np.linspace(dTidr_ib_wall, dTidr_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        dTidr_ob = np.linspace(dTidr_sep_cut[-1], dTidr_ob_wall, inp.xi_ob_pts, endpoint=True)

        dTedr_ib_wall = dTedr_sep_cut[0]
        dTedr_ob_wall = dTedr_sep_cut[-1]
        dTedr_ib = np.linspace(dTedr_ib_wall, dTedr_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        dTedr_ob = np.linspace(dTedr_sep_cut[-1], dTedr_ob_wall, inp.xi_ob_pts, endpoint=True)

        BT_ib_wall = BT_sep_cut[0]
        BT_ob_wall = BT_sep_cut[-1]
        BT_ib = np.linspace(BT_ib_wall, BT_sep_cut[0], inp.xi_ib_pts, endpoint=False)
        BT_ob = np.linspace(BT_sep_cut[-1], BT_ob_wall, inp.xi_ob_pts, endpoint=True)

        # combine inboard, seperatrix, and outboard points. These now comprise all the values in the xi direction.
        ni_xi = np.concatenate((ni_ib, ni_sep_cut, ni_ob))
        ne_xi = np.concatenate((ne_ib, ne_sep_cut, ne_ob))
        Ti_xi = np.concatenate((Ti_ib, Ti_sep_cut, Ti_ob))
        Te_xi = np.concatenate((Te_ib, Te_sep_cut, Te_ob))
        dnidr_xi = np.concatenate((dnidr_ib, dnidr_sep_cut, dnidr_ob))
        dnedr_xi = np.concatenate((dnedr_ib, dnedr_sep_cut, dnedr_ob))
        dTidr_xi = np.concatenate((dTidr_ib, dTidr_sep_cut, dTidr_ob))
        dTedr_xi = np.concatenate((dTedr_ib, dTedr_sep_cut, dTedr_ob))
        BT_xi = np.concatenate((BT_ib, BT_sep_cut, BT_ob))

        ib_leg_length = core.lines.div.ib.length
        ob_leg_length = core.lines.div.ob.length
        sep_length = core.lines.sep_closed.length
        ib_frac = ib_leg_length / (ib_leg_length + sep_length + ob_leg_length)
        sep_frac = sep_length / (ib_leg_length + sep_length + ob_leg_length)
        ob_frac = ob_leg_length / (ib_leg_length + sep_length + ob_leg_length)

        # specify the points along xi
        xi_ib_div = np.linspace(0,
                                frac_to_remove,
                                inp.xi_ib_pts,
                                endpoint=False)

        xi_sep = np.linspace(frac_to_remove,
                             1.0 - frac_to_remove,
                             len(dnidr_sep_cut),
                             endpoint=False)

        xi_ob_div = np.linspace(1.0 - frac_to_remove,
                                1.0,
                                inp.xi_ob_pts,
                                endpoint=True)

        xi_pts = np.concatenate((xi_ib_div, xi_sep, xi_ob_div))

        # model perpendicular particle and heat transport using Bohm Diffusion
        D_perp = Ti_xi / (16.0 * elementary_charge * BT_xi)
        Chi_perp = 5.0 * Ti_xi / (32.0 * elementary_charge * BT_xi)

        Gamma_perp = -D_perp * dnidr_xi
        Q_perp = -ni_xi * Chi_perp * dTidr_xi - \
                 3.0 * Ti_xi * D_perp * dnidr_xi

        delta_sol_n = D_perp * ni_xi / Gamma_perp
        delta_sol_T = Chi_perp / (Q_perp / (ni_xi * Ti_xi) - 3.0 * D_perp / delta_sol_n)
        delta_sol_E = 2 / 7 * delta_sol_T

        # now calculate densities and temperatures radially outward from the seperatrix for a distance
        # long enough that the wall is enclosed, so we can get densities and temperatures along the wall

        # first draw wall line through 2d strip model to get n, T along the line
        # we do this first so we can find the farthest point and make sure that we
        # make our SOL strip wide enough to go all the way to even the farthest point
        # on the wall
        wall_pts = np.asarray(inp.wall_line.xy).T
        # ib_int_pt = np.asarray(core.lines.div.ib.intersection(inp.wall_line).xy).T
        # ob_int_pt = core.lines.div.ob.intersection(inp.wall_line)
        ib_int_pt = np.asarray(core.pts.strike.ib)
        ob_int_pt = np.asarray(core.pts.strike.ob)

        wall_start_pos = np.where((wall_pts == ib_int_pt).all(axis=1))[0][0]
        wall_line_rolled = LineString(np.roll(wall_pts, -wall_start_pos, axis=0))

        wall_line_cut = cut(wall_line_rolled,
                            wall_line_rolled.project(Point(ob_int_pt), normalized=True))[0]

        # add points to wall line for the purpose of getting n, T along the wall. These points
        # won't be added to the main wall line or included in the triangulation.
        # for i, v in enumerate(np.linspace(0, 1, 300)):
        #    #interpolate along wall_line_cut to find point to add
        #    pt = wall_line_cut.interpolate(v, normalized=True)
        #    #add point to wall_line_cut
        #    union = wall_line_cut.union(pt)
        #    result = [geom for geom in polygonize(union)][0]
        #    wall_line_cut = LineString(result.exterior.coords)

        wall_nT_pts = np.asarray(wall_line_cut)
        num_wall_pts = len(wall_nT_pts)
        wall_pos_norm = np.zeros(num_wall_pts)
        wall_dist = np.zeros(num_wall_pts)

        for i, pt in enumerate(wall_nT_pts):
            wall_pt = Point(pt)
            sep_pt_pos = core.lines.ib2ob.project(Point(wall_pt), normalized=True)
            sep_pt = core.lines.ib2ob.interpolate(sep_pt_pos, normalized=True)
            wall_pos_norm[i] = wall_line_cut.project(wall_pt, normalized=True)
            wall_dist[i] = wall_pt.distance(sep_pt)

        r_max = np.amax(wall_dist) * 1.1  # the 1.1 multiplication just ensures that we go a bit farther than necessary
        twoptdiv_r_pts = 20

        r_pts = np.linspace(0, r_max, twoptdiv_r_pts)
        xi, r = np.meshgrid(xi_pts, r_pts)
        sol_ni = ni_xi * np.exp(-r / delta_sol_n)

        sol_ne = ne_xi * np.exp(-r / delta_sol_n)
        sol_Ti = Ti_xi * np.exp(-r / delta_sol_T)
        sol_Te = Te_xi * np.exp(-r / delta_sol_T)

        # set some minimum values
        min_n = 1E15
        min_T = 2.0 * 1.6021E-19
        sol_ni = np.where(sol_ni < min_n, min_n, sol_ni)
        sol_ne = np.where(sol_ne < min_n, min_n, sol_ne)
        sol_Ti = np.where(sol_Ti < min_T, min_T, sol_Ti)
        sol_Te = np.where(sol_Te < min_T, min_T, sol_Te)

        # draw sol lines through 2d strip model to get n, T along the lines
        sol_line_dist = np.zeros((len(xi_pts), len(self.sol_lines_cut)))
        sol_nT_pts = np.zeros((len(xi_pts), 2, len(self.sol_lines_cut)))
        for i, sol_line in enumerate(self.sol_lines_cut):
            for j, xi_val in enumerate(xi_pts):
                sol_pt = sol_line.interpolate(xi_val, normalized=True)
                sol_nT_pts[j, :, i] = np.asarray(sol_pt.xy).T
                sep_pt_pos = core.lines.ib2ob.project(sol_pt, normalized=True)
                sep_pt = core.lines.ib2ob.interpolate(sep_pt_pos, normalized=True)
                sol_line_dist[j, i] = sol_pt.distance(sep_pt)

        sol_line_ni = np.zeros((len(xi_pts), len(self.sol_lines_cut)))
        sol_line_ne = np.zeros((len(xi_pts), len(self.sol_lines_cut)))
        sol_line_Ti = np.zeros((len(xi_pts), len(self.sol_lines_cut)))
        sol_line_Te = np.zeros((len(xi_pts), len(self.sol_lines_cut)))

        pts_ni_sol = np.zeros((0, 3))
        pts_ne_sol = np.zeros((0, 3))
        pts_Ti_sol = np.zeros((0, 3))
        pts_Te_sol = np.zeros((0, 3))
        for i, sol_line in enumerate(self.sol_lines_cut):
            sol_line_ni[:, i] = griddata(np.column_stack((xi.flatten(), r.flatten())),
                                         sol_ni.flatten(),
                                         np.column_stack((np.linspace(0, 1, len(xi_pts)), sol_line_dist[:, i])),
                                         method='linear')
            sol_line_ne[:, i] = griddata(np.column_stack((xi.flatten(), r.flatten())),
                                         sol_ne.flatten(),
                                         np.column_stack((np.linspace(0, 1, len(xi_pts)), sol_line_dist[:, i])),
                                         method='linear')
            sol_line_Ti[:, i] = griddata(np.column_stack((xi.flatten(), r.flatten())),
                                         sol_Ti.flatten(),
                                         np.column_stack((np.linspace(0, 1, len(xi_pts)), sol_line_dist[:, i])),
                                         method='linear')
            sol_line_Te[:, i] = griddata(np.column_stack((xi.flatten(), r.flatten())),
                                         sol_Te.flatten(),
                                         np.column_stack((np.linspace(0, 1, len(xi_pts)), sol_line_dist[:, i])),
                                         method='linear')

            # append to SOL n, T arrays
            pts_ni_sol = np.vstack((pts_ni_sol, np.column_stack((sol_nT_pts[:, :, i], sol_line_ni[:, i]))))
            pts_ne_sol = np.vstack((pts_ne_sol, np.column_stack((sol_nT_pts[:, :, i], sol_line_ne[:, i]))))
            pts_Ti_sol = np.vstack((pts_Ti_sol, np.column_stack(
                (sol_nT_pts[:, :, i], sol_line_Ti[:, i] / 1.0E3 / 1.6021E-19))))  # converting back to kev
            pts_Te_sol = np.vstack((pts_Te_sol, np.column_stack(
                (sol_nT_pts[:, :, i], sol_line_Te[:, i] / 1.0E3 / 1.6021E-19))))  # converting back to kev

        self.sol_nT = namedtuple('sol_nT', 'ni ne Ti Te')(pts_ni_sol, pts_ne_sol, pts_Ti_sol, pts_Te_sol)

        wall_ni = griddata(np.column_stack((xi.flatten(), r.flatten())),
                           sol_ni.flatten(),
                           np.column_stack((wall_pos_norm, wall_dist)),
                           method='linear')
        wall_ne = griddata(np.column_stack((xi.flatten(), r.flatten())),
                           sol_ne.flatten(),
                           np.column_stack((wall_pos_norm, wall_dist)),
                           method='linear')
        wall_Ti = griddata(np.column_stack((xi.flatten(), r.flatten())),
                           sol_Ti.flatten(),
                           np.column_stack((wall_pos_norm, wall_dist)),
                           method='linear')
        wall_Te = griddata(np.column_stack((xi.flatten(), r.flatten())),
                           sol_Te.flatten(),
                           np.column_stack((wall_pos_norm, wall_dist)),
                           method='linear')

        # append to master arrays
        pts_ni_wall = np.column_stack((wall_nT_pts, wall_ni))
        pts_ne_wall = np.column_stack((wall_nT_pts, wall_ne))
        pts_Ti_wall = np.column_stack((wall_nT_pts, wall_Ti / 1.0E3 / 1.6021E-19))  # in kev
        pts_Te_wall = np.column_stack((wall_nT_pts, wall_Te / 1.0E3 / 1.6021E-19))  # in kev

        self.wall_nT = namedtuple('wall_nT', 'ni ne Ti Te')(pts_ni_wall, pts_ne_wall, pts_Ti_wall, pts_Te_wall)

        # # uncomment this if you're debugging
        # plt.contourf(xi, r, sol_Ti, 500)
        # plt.colorbar()
        # for i, v in enumerate(self.sol_lines_cut):
        #     plt.plot(xi_pts, sol_line_dist[:, i])
        # plt.plot(np.linspace(0, 1, num_wall_pts), wall_dist, color='black')
        # plt.show()
        # sys.exit()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import griddata

def calc_psi_norm(R, Z, psi, xpt, axis_mag):
    # normalize psi
    # psi_interp = Rbf(R, Z, psi)
    # psi_min = psi_interp(axis_mag[0], axis_mag[1])
    #
    # psi_shifted = psi - psi_min  # set center to zero
    # psi_shifted_interp = Rbf(R, Z, psi_shifted)
    # psi_shifted_xpt = psi_shifted_interp(xpt[0], xpt[1])
    #
    # psi_norm = psi_shifted / psi_shifted_xpt

    psi_min = griddata(np.column_stack((R.flatten(), Z.flatten())),
                       psi.flatten(),
                       [axis_mag[0], axis_mag[1]],
                       method='cubic')

    psi_shifted = psi - psi_min  # set center to zero
    psi_shifted_xpt_l, psi_shifted_xpt_u = None, None

    if xpt[0] is not None:

        psi_shifted_xpt_l = griddata(np.column_stack((R.flatten(), Z.flatten())),
                               psi_shifted.flatten(),
                               [xpt[0][0], xpt[0][1]],
                               method='cubic')
    if xpt[1] is not None:
        psi_shifted_xpt_u = griddata(np.column_stack((R.flatten(), Z.flatten())),
                               psi_shifted.flatten(),
                               [xpt[1][0], xpt[1][1]],
                               method='cubic')
    psi_shifted_xpt = [psi_shifted_xpt_l, psi_shifted_xpt_u]
    if xpt[1] is None:
        psi_norm = psi_shifted / np.average(psi_shifted_xpt_l)
    elif xpt[0] is None:
        psi_norm = psi_shifted / np.average(psi_shifted_xpt_u)
    else:
        psi_norm = psi_shifted / np.average(psi_shifted_xpt)

    return psi_norm



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from shapely.geometry import Point, LineString

def cut(line, distance):
    # Cuts a line in two at a distance from its starting point
    if distance <= 0.0 or distance >= 1.0:
        return [LineString(line)]
    coords = list(line.coords)
    for i, p in enumerate(coords):
        pd = line.project(Point(p), normalized=True)
        if pd == distance:
            return [
                LineString(coords[:i+1]),
                LineString(coords[i:])]
        if pd > distance:
            cp = line.interpolate(distance, normalized=True)
            return [
                LineString(coords[:i] + [(cp.x, cp.y)]),
                LineString([(cp.x, cp.y)] + coords[i:])]




#!/usr/bin/python2.7

import numpy as np
import os
import texttable
import sys
from shapely.geometry import LineString
import matplotlib.pyplot as plt


def getGT3Test(iolFlag=True, verbose=False, neutFlag=True, mode="coreonly"):
    try:
        from GT3 import gt3
    except ImportError:
        ImportError("Unable to load GT3 module")
    return gt3(preparedInput=TestClass(), iolFlag=iolFlag, verbose=verbose, neutFlag=neutFlag, mode=mode)

class TestClass:

    def __init__(self, shotid=164436, timeid=3740):
        self.shotid = shotid
        self.timeid = timeid
        print "This is the GT3 Test Base Class"
        print "This class uses DIII-D shot %s.%s for debugging purposes" % (self.shotid, self.timeid)
        print "Use the print_summary method to see a summary of this shot"

        """
        Generate various 0D parameters for the plasma from shot 164436.3740
        """
        self.iolFlag = True
        self.BT0 = 2.004
        self.wall_ni_min = 1.0E15
        self.wall_ne_min = 1.0E15
        self.wall_Ti_min = 0.005
        self.wall_Te_min = 0.005
        self.pfr_ni_val = 1.0E14
        self.pfr_ne_val = 1.0E14
        self.pfr_Ti_val = 0.002
        self.pfr_Te_val = 0.002
        self.ebeam = 64.72
        self.abeam = 2
        self.alphain = 0.6475
        self.pbeam = 0.8
        self.rtang = 1.146
        self.R_loss = .5
        self.sep_val = 1.0


        """
        Define the meshing parameters
        """
        self.thetapts_approx = 50
        self.edge_rho = 0.7
        self.rhopts_edge = 100
        self.rhopts_core = 10
        self.core_thetapts_ntrl = 50
        self.edge_rho_ntrl = 0.8
        self.rhopts_edge_ntrl = 5
        self.rhopts_core_ntrl = 10
        self.sollines_psi_max = 1.07
        self.num_sollines = 3
        self.xi_ib_pts = 10
        self.xi_ob_pts = 10
        self.numcos = 20
        self.sep_val = 1.0

        """
        Define the meshing parameters
        """

        self.er_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_er.dat' % (self.shotid, self.timeid), comments='#')
        self.ne_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_ne.dat' % (self.shotid, self.timeid), comments='#')
        self.nD_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_nd.dat' % (self.shotid, self.timeid), comments='#')
        self.nC_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_nc.dat' % (self.shotid, self.timeid), comments='#')
        self.Te_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_Te.dat' % (self.shotid, self.timeid), comments='#')
        self.Ti_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_Ti.dat' % (self.shotid, self.timeid), comments='#')
        self.vpolC_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_vpolC.dat' % (self.shotid, self.timeid), comments='#')
        self.vpolD_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_vpolD.dat' % (self.shotid, self.timeid), comments='#')
        self.vtorD_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_vtorD.dat' % (self.shotid, self.timeid), comments='#')
        self.vtorC_data = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_vtorC.dat' % (self.shotid, self.timeid), comments='#')
        self.psirz_exp = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_psirz.txt' % (self.shotid, self.timeid), comments='#')
        self.wall_exp = np.genfromtxt(os.path.dirname(__file__) + '/TestBaseProfiles/gt3_diiid_wall.dat', comments='#')
        self.wall_line = LineString(self.wall_exp)
        self.nbeams_loc = os.path.dirname(__file__) + '/../nbeams/bin/Release/nbeams'
        self.beams_json = os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_beams.json' % (self.shotid, self.timeid)
        self.beams_out_json = os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_outbeams.json' % (self.shotid, self.timeid)
        self.neutfile_loc = os.path.dirname(__file__) + '/TestBaseProfiles/gt3_%s_%s_outneuts.json' % (self.shotid, self.timeid)

    def print_summary(self):
        print "0D Parameters \n"
        table1 = texttable.Texttable()
        table1.set_cols_align(["l", "l", "m", "m"])
        table1.set_cols_valign(["m", "m", "m", "m"])
        table1.add_rows([["Variable name", "Description", "Value", "Units"],
                        ["shotid", "DIII-D Shot ID", self.shotid, "n/a"],
                        ["timeid", "DIII-D Time ID", self.timeid, "n/a"],
                        ["BT0", "Toroidal magnetic field at rho=0", self.BT0, "T"],
                        ["wall_ni_min", "Minimum wall ion density", self.wall_ni_min, "#/m^3"],
                        ["wall_ne_min", "Minimum wall electron density", self.wall_ne_min, "#/m^3"],
                        ["wall_Ti_min", "Minimum wall ion temperature", self.wall_Ti_min, "keV"],
                        ["wall_Te_min", "Minimum wall electron temperature", self.wall_Te_min, "keV"],
                        ["pfr_ni_val", "Ion density in the private flux region", self.pfr_ni_val, "#/m^3"],
                        ["pfr_ne_val", "Electron density in the private flux region", self.pfr_ne_val, "#/m^3"],
                        ["pfr_ti_val", "Ion temperature in the private flux region", self.pfr_Ti_val, "keV"],
                        ["pfr_te_val", "Electron temperature in the private flux region", self.pfr_Te_val, "keV"],
                        ["ebeam", "Power-averaged NBI beam energy", self.ebeam, "keV"],
                        ["abeam", "Some value for NBeams",  self.abeam, "1"],
                        ["alphain", "Power-averaged launch angle (maybe)", self.alphain, "??"],
                        ["pbeam", "Duty-cycle-averaged beam power", self.pbeam, "MW"],
                        ["rtang", "Radius of tangency of beam ions", self.rtang, "m"],
                        ["R_loss", "Fraction of IOL particles that do not return on loss orbits", self.R_loss, "1"]
                        ])

        print table1.draw() + "\n"

        print "1D Parameters \n"
        table2 = texttable.Texttable()
        table2.set_cols_align(["l", "l", "c"])
        table2.set_cols_valign(["m", "m", "m"])
        table2.add_rows([["Variable name", "Description", "Units"],
                         ["Er", "Radial electric field", "V/m"],
                         ["ne", "Electron density", "#/m^3"],
                         ["nD", "Main ion density", "#/m^3"],
                         ["nC", "Carbon density", "#/m^3"],
                         ["Te", "Electron temperature", "keV"],
                         ["Ti", "Ion temperature", "keV"],
                         ["VpolC", "Poloidal carbon velocity", "m/s"],
                         ["VpolD", "Poloidal main ion velocity", "m/s"],
                         ["VtorC", "Toroidal carbon velocity", "m/s"],
                         ["VtorD", "Toroidal main ion velocity", "m/s"],
                         ["wall_line", "LineString object for the machine wall", "n/a"]
                         ])
        print table2.draw() + "\n"

        print "Computation-related Parameters \n"
        table3 = texttable.Texttable()
        table3.set_cols_align(["l", "l", "m"])
        table3.set_cols_valign(["m", "m", "m"])
        table3.add_rows([["Variable name", "Description", "Value"],
                         ["thetapts_approx", "Approximate number of theta points (may be modified in situ)", self.thetapts_approx],
                         ["edge_rho", "Definition of the plasma edge", self.edge_rho],
                         ["rhopts_edge", "Number of rho values in the edge", self.rhopts_edge],
                         ["rhopts_core", "Number of rho values in the core", self.rhopts_core],
                         ["core_thetapts_ntrl", "Number of theta values in the neutrals core calculation", self.core_thetapts_ntrl],
                         ["edge_rho_ntrl", "Definition of the edge in the neutrals calculation", self.edge_rho_ntrl],
                         ["rhopts_edge_ntrl", "Number of rho values in the edge in the neutrals calculation", self.rhopts_edge_ntrl],
                         ["rhopts_core_ntrl", "Number of rho values in the core in the neutrals calculation", self.rhopts_core_ntrl],
                         ["sollines_psi_max", "Maximum psi value after the scrape-off layer", self.sollines_psi_max],
                         ["num_sollines", "Number of lines after the scrape-off layer", self.num_sollines],
                         ["numcos", "Number of cosine angles used in IOL calculation", self.numcos]
                         ])
        print table3.draw()

        print "Nbeams location: %s" % self.nbeams_loc

if __name__ == "__main__":

    plasma = getGT3Test()
    plasma.run_NBI()
    print "DONE"



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np

def calc_cxcool(core, n, T):
    """Calculates charge exchange cooling with the slow neutrals"""
    slowNeutDens = np.array(map(lambda x: n.n.s[-1] * x ** 15, core.r[:, 0] / core.a))
    totalNeutDens = np.array(map(lambda x: n.n.tot[-1] * x ** 15, core.r[:, 0] / core.a))
    #result = 1.5 * n.i * T.i.J * slowNeutDens* ((core.sv.el.st[:,0] + core.sv.el.st[:,0]* (totalNeutDens/n.i)) + core.sv.cx.st[:,0])
    result = 1.5 * n.i * T.i.J * slowNeutDens * ((core.sv.el.st[:, 0] + core.sv.cx.st[:, 0])) / 1E4 # TODO: Work this out for n-n elastic scattering
    return result


#/usr/bin/python

from GT3 import gt3
import matplotlib.pyplot as plt


def _plot_base(rho, val, xLabel=r'$\rho$', yLabel="Value", title="Title", color='red', edge=False):

    plot = plt.figure()
    fig = plot.add_subplot(111)
    fig.set_xlabel(xLabel, fontsize=30)
    fig.set_ylabel(yLabel, fontsize=30)
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    fig.set_title(title)
    if edge:
        fig.set_xlim(0.85, 1.0)
    fig.scatter(rho, val, color=color, s=8)
    plt.show()
    return fig


class IOLComparison:

    def __init__(self, filename):

        self.plasmaWithIOL = gt3(inputFile=filename)
        self.plasmaWithIOL.run_radial_transport()

        self.plasmaWithoutIOL = gt3(inputFile=filename)
        self.plasmaWithoutIOL.disable_IOL()
        self.plasmaWithoutIOL.run_radial_transport()

    def comp_gamma_diff(self, edge=True):
        fig = _plot_base(self.plasmaWithIOL.rtrans.rhor,
                         self.plasmaWithIOL.rtrans.gamma_diff_D,
                         xLabel=r'$\rho$',
                         yLabel=r"$\Gamma_{i,r}[\#/m^2 s]$",
                         title="Radial particle flux w/ & w/out IOL",
                         color="red", edge=edge)
        fig.scatter(self.plasmaWithoutIOL.rtrans.rhor,
                    self.plasmaWithoutIOL.rtrans.gamma_diff_D,
                    color="blue", s=8)
        fig.legend(["IOL Corrected", "Uncorrected"])
        return fig

    def comp_D(self, edge=True):
        fig = _plot_base(self.plasmaWithIOL.rtrans.rhor,
                         self.plasmaWithIOL.rtrans.D_i,
                         xLabel=r'$\rho$',
                         yLabel=r"$D{i,r}[m/s^2]$",
                         title="Radial Diffusion Coefficient w/ & w/out IOL",
                         color="red", edge=edge)
        fig.scatter(self.plasmaWithoutIOL.rtrans.rhor,
                    self.plasmaWithoutIOL.rtrans.D_i,
                    color="blue", s=8)
        fig.legend(["IOL Corrected", "Uncorrected"])
        return fig



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from collections import namedtuple
import numpy as np

class debugCore():

    def __init__(self):

        rho1d= np.concatenate((np.linspace(0, 0.8, 10, endpoint=False), np.linspace(0.8, 1, 100, endpoint=True)))
        self.a = 0.5640316391827651
        self.kappa = namedtuple('kappa', 'axis sep')(1.2507812426138631,1.6533128554966692)
        self.shaf_shift = 0.04273763321406595
        self.theta, self.rho = np.meshgrid([0., 1.], rho1d)
        loader=np.loadtxt("inputs/nbDebugProfs.dat")
        # This named tuple is generated by:
        #
        # np.savetxt('nbDebugProfs.dat', (core.n.e[:,0] / ((1.+.025*6.0)), core.n.e[:, 0], core.T.i.kev[:, 0], core.T.e.kev[:, 0]))
        #
        # in a module supporting calls to core. 9/4/2018 nbeams debug profiles obtained from shot 118890.1560

        self.n = namedtuple('n','i e')(np.array([loader[0],loader[0]]).transpose(),np.array([loader[1],loader[1]]).transpose())
        self.T = namedtuple('T','i e')(namedtuple('i', 'kev')(np.array([loader[2],loader[2]]).transpose()),
                                       namedtuple('e', 'kev')(np.array([loader[3],loader[3]]).transpose()))


        self.pts = namedtuple('pts','axis')(namedtuple('axis','mag')([1.7151768587049954,None]))
        self.kappa_vals = namedtuple('kappa_vals', 'axis sep')(1.2507812426138631, 1.6533128554966692)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from scipy.interpolate import UnivariateSpline


def balance(gamma, interm, r, sa, x):
    return r[x], UnivariateSpline(r, interm, k=3, s=0).integral(0., r[x]), gamma[x]*sa(r[x])



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from scipy.interpolate import UnivariateSpline
import numpy as np
from collections import namedtuple
from scipy import constants
from math import sqrt


m_d = constants.physical_constants['deuteron mass'][0]
e = constants.elementary_charge
z_d = 1  # atomic number of deuterium
z_c = 6  # atomic number of carbon
ch_d = e * z_d  # charge of deuterium
ch_c = e * z_c  # charge of carbon



class Chi:
    """
    Chi class claculates various chis and provides source terms as necessary
    """
    def __init__(self, data, core, n, T, L, reInterp=False):

        self.Qi = data.Qi_diff
        self.Qe = data.Qe_diff
        self.conv15 = UnivariateSpline(core.r[:,0], .5 * ch_d * data.gamma_diff_D * T.i.ev, k=3, s=0)(core.r[:,0])
        self.conv25 = UnivariateSpline(core.r[:,0], .5 * ch_d * data.gamma_diff_D * T.i.ev, k=3, s=0)(core.r[:,0])
        #self.heatvisc = self.viscCalc(data, core, n, T)
        self.heatvisc = np.zeros(self.conv25.shape)
        self.heatin = UnivariateSpline(core.r[:,0], .5 * data.gamma_diff_D * m_d * (data.vtor_D_total**2 + data.vpol_D**2), k=3, s=0)(core.r[:,0]) # TODO: Provide logic that uses vtor_D_intrin/fluid depending on IOL Switch, currently too small to matter
        if reInterp:
            L_T_i = UnivariateSpline(core.r[:,0], L.T.i, k=3, s=0)(core.r[:,0])
            T_i_ev = UnivariateSpline(core.r[:,0], T.i.ev, k=3, s=0)(core.r[:,0])
            n_i = UnivariateSpline(core.r[:,0], n.i, k=3, s=0)(core.r[:,0])
            self.chi = namedtuple('chi','i e')(
                namedtuple('i','chi1 chi2 chi3 chi4')(
                    UnivariateSpline(core.r[:,0], (self.Qi) * L_T_i / (n_i * T_i_ev * ch_d), k=1, s=5)(core.r[:,0]),
                    UnivariateSpline(core.r[:,0], (self.Qi - self.conv25) * L_T_i / (n_i * T_i_ev * ch_d), k=1, s=5)(core.r[:,0]),
                    UnivariateSpline(core.r[:,0], ((self.Qi - self.conv25 - self.heatin) * L_T_i / (n_i * T_i_ev * ch_d)), k=1, s=5)(core.r[:,0]),
                    UnivariateSpline(core.r[:,0], ((self.Qi - self.conv25 - self.heatin * self.heatvisc) * L_T_i / (n_i * T_i_ev * ch_d)), k=1, s=5)(core.r[:,0])
                ),self.calc_chi_e(data, n, L, T)
            )
        else:
            self.chi = namedtuple('chi','i e')(
                namedtuple('i','chi1 chi2 chi3 chi4')(
                    (self.Qi) * L.T.i / (n.i * T.i.ev * ch_d),
                    (self.Qi - self.conv25) * L.T.i / (n.i * T.i.ev * ch_d),
                    (self.Qi - self.conv25 - self.heatin) * L.T.i / (n.i * T.i.ev * ch_d),
                    (self.Qi - self.conv25 - self.heatin * self.heatvisc) * L.T.i / (n.i * T.i.ev * ch_d)
                ),self.calc_chi_e(data, n, L, T)
            )


    def calc_chi_e(self, data, n, L, T):
        gameltemp = 1.0* data.gamma_diff_D + 6.0 * data.gamma_C   #OHHHH gamma electron!!! HA HA HA HA WE DON'T KNOW WHAT THE FUCK GAMMA_C IS

        return L.T.e * ((self.Qe / (ch_d * n.e * T.e.ev)) - 2.5 * gameltemp / n.e)

    def viscCalc(self, data, core, n, T):
        f1 = 1 # Must determine later, appears to be geometeric factor in Eq(4) in POP052504(2010)
        fp = core.B_p_fsa/core.B_t_fsa
        vth = [sqrt(2. * a  * ch_d / m_d) for a in T.i.ev]
        eta0 = [a * m_d * b * c * core.R0_a * f1 for a,b,c in zip(n.i, vth, core.q[:,0])]
        eta4 = [a * m_d * c * ch_d / (ch_d * abs(b)) for a, b, c in zip(n.i, core.B_t_fsa, T.i.ev)]
        vrad = data.gamma_diff_D / n.i

        # a = vtor    b = fp   c = eta0
        # d = vrad    f = vthet g = eta 4

        return [a * (b * c * d - .5 * g * (4.0 * a + f)) - .5 * f * (c * d + g * (a + .5 * f)) for a, b,c,d,f,g in zip(data.vtor_D_total, fp, eta0, vrad, data.vpol_D, eta4)]




#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from GT3.RadialTransport.Functions.CalcTorque import calc_torque
from math import cos

def calc_mom_src_nbi(beam,  n, z_eff, R0_a, fforb):
    """Calculates toroidal momentum input from a neutral beam (or beam component)

    :param fforb:
    :param beam:
    :param n:
    :param z_eff:
    :param rpts:
    :param R0_a:
    :return:
    """

    def calc_atten(beamE, n, z_eff, rpts=len(n.i)):  # TODO: Figure out what this is. has to do with beam attenuation
        delma = 1 / (rpts - 1)
        alphain = 0.6475  # TODO: no idea what this is
        xlam = 5.5E17 * (beamE / 1) / (2.0 * 0.5 * n.i * z_eff ** 0.5)
        atten = 1 - np.exp(-delma / cos(alphain) / xlam)
        return atten

#
#   Note: Legacy GTEDGE had the beam energy in keV I THINK. Using Joules just gives 0 torque.
#
    try:
        beam.E.J[1] #checks to see if multiple beams
        atten, unatten, torque, tor_mom = [], [], [], []
        for i in range(len(beam.E.J)):
            atten.append(calc_atten(beam.E.kev[i] , n, z_eff))
            unatten.append(1-atten[i])
            torque.append(calc_torque(beam, fforb, index=i))
            if beam.coI[i]:
                tor_mom.append(torque[i]*unatten*atten/R0_a)
            else:
                tor_mom.append(-1 * torque[i] * atten * unatten / R0_a)

        return np.asarray([sum(x) for x in zip(*tor_mom)][0])
    except:
        atten = calc_atten(beam.E.kev, n, z_eff)
        unatten = 1-atten  # might need to flip atten...
        torque = calc_torque(beam, fforb)

        tor_mom = unatten*atten*torque/R0_a
    return tor_mom


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from scipy.constants import constants

E_phi = 0.04  # toroidal electrostatic potential

e = constants.elementary_charge

def calc_mbal_rhs(mom_src_ext, z, n, B_p, gamma):
    """ The function formerly known as y11 and y22
    """

    mbal_rhs = mom_src_ext + (z * e) * (n * E_phi + B_p * gamma)

    return mbal_rhs


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

def writeFile(s, t, fpath, data, reNeut=False):

    """
    This function writes an input file and saves it in the inputs/shot_timeid folder.

    :param s: shot id
    :param t: time id
    :param fpath: This is the file that is to be written. See the folder structure in getVals() for file structure
    :param data: These data are from the interactive getVals() code that lets you interactively give parameters for
                 setting up input files
    :param reNeut: This flag tells GT3 whether a new run of neutpy needs to be performed. Concerns ntrl_switch in gt3
    :return: Your mom
    """
    with open(fpath, "w") as f:
        f.write("d3d_iter = 1 \n")
        f.write("\n")
        f.write("#1D PROFILE INPUT FILES \n")
        f.write("ne_file         = %s_%s/gt3_%s_%s_ne.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("nD_file         = %s_%s/gt3_%s_%s_nD.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("Te_file         = %s_%s/gt3_%s_%s_Te.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("Ti_file         = %s_%s/gt3_%s_%s_Ti.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("er_file         = %s_%s/gt3_%s_%s_er.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("fz1_file        = %s_%s/gt3_%s_%s_fz1.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("fracz_file      = %s_%s/gt3_%s_%s_fracz.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("exlti_file      = %s_%s/gt3_%s_%s_exlti.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("exlte_file      = %s_%s/gt3_%s_%s_exlte.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("exlni_file      = %s_%s/gt3_%s_%s_exlni.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("vpolC_file      = %s_%s/gt3_%s_%s_vpolC.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("vtorC_file      = %s_%s/gt3_%s_%s_vtorC.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("vpolD_file      = %s_%s/gt3_%s_%s_vpolD.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("vtorD_file      = %s_%s/gt3_%s_%s_vtorD.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("q_file          = %s_%s/gt3_%s_%s_q.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("zbar2_file      = %s_%s/gt3_%s_%s_zbar2.dat    \n" % (str(s), str(t), str(s), str(t)))
        f.write("\n")
        f.write("2D QUANTITIES INPUT FILES \n")
        f.write("bpol_file       = %s_%s/gt3_%s_%s_bpol.dat  \n" % (str(s), str(t), str(s), str(t)))
        f.write("btor_file       = %s_%s/gt3_%s_%s_btor.dat  \n" % (str(s), str(t), str(s), str(t)))
        f.write("psirz_file      = %s_%s/gt3_%s_%s_psirz.dat \n" % (str(s), str(t), str(s), str(t)))
        f.write("\n")
        f.write("2D LINE INPUT FILES \n")
        f.write("wall_file       = %s_%s/gt3_diiid_wall.dat \n" % (str(s), str(t)))
        f.write("\n")

        f.write("# CONSTANTS \n")
        f.write("xk = %s \n" % str(data.xk))
        f.write("delma = %s \n" % str(data.delma))
        f.write("xnuioni = 0.0 \n")
        f.write("xnuati = 0.0 \n")
        f.write("\n")

        f.write("#NEUTRAL BEAM DEPOSITION \n")
        f.write(
            "nbeams_loc = /home/jonathan/Dropbox/GTEDGE/MyPrograms/GTEDGE/lib/beams/NBeamsMDS/NBeams/bin/Release/nbeams \n")
        f.write("ebeam = %s \n" % str(data.ebeam))
        f.write("abeam = 2 \n")
        f.write("alphain = .6475 \n")
        f.write("pbeam = %s \n" % str(data.pbeam))
        f.write("rtang = %s \n" % str(data.rtang))
        f.write("bknot = %s \n" % str(data.bknot))
        f.write("pwrfrac1 = %s \n" % str(data.pwrfrac1))
        f.write("pwrfrac2 = %s \n" % str(data.pwrfrac2))
        f.write("pwrfrac3 = %s \n" % str(data.pwrfrac3))
        f.write("epsknot = %s \n" % str(data.epsknot))
        f.write("eps_sep = %s \n" % str(data.epssep))
        f.write("shftknot = %s \n" % str(data.shftknot))
        f.write("\n")

        f.write("#GENERAL GEOMETRY \n")
        f.write("a = %s \n" % str(data.aminor))
        f.write("BT0 = %s \n" % str(data.bphi * -1.))
        f.write("Z0 = 0.0 \n")
        f.write("thetapts_approx = %s \n" % str(int(data.thetapts)))
        f.write("rhopts = %s \n" % str(201))
        f.write("edge_rho = 0.8 \n")
        try:
            f.write("rhopts_edge = %s \n" % str(data.rhopts_edge))
        except:
            f.write("rhopts_edge = %s \n" % str(100))
        try:
            f.write("rhopts_core = %s \n" % str(data.rhopts_core))
        except:
            f.write("rhopts_core = %s \n" % str(10))
        try:
            f.write("thetapts_approx = %s \n" % str(data.thetapts))
        except:
            f.write("thetapts_approx = %s \n" % str(30))
        f.write("\n")
        f.write("#NEUTRALS CALCULATION \n")
        if reNeut:
            f.write("ntrl_switch = 2 \n")
        elif reNeut == False:
            f.write("ntrl_switch = 1 \n")
        else:
            raise Exception("reNeut not defined")
        f.write("edge_rho_ntrl = 0.8 \n")
        f.write("neut_outfile =  inputs/%s_%s/gt3_%s_%s_neut.dat \n" % (str(s), str(t), str(s), str(t)))
        #        f.write("neut_outfile = gt3_%s_%s_neut.dat \n" % (str(s),str(t)))
        f.write("rhopts_edge_ntrl = %s \n" % str(10))
        f.write("ntrl_thetapts = 33 \n")
        f.write("\n")

        f.write("#BACKGROUND DENSITIES AND TEMPERATURES (IF NOT READING FROM INPUT FILE) \n")
        f.write("#ni0 = 3.629E19 \n")
        f.write("#ni9 = 1.523E19\n")
        f.write("#ni_sep = 0.3E19\n")
        f.write("#ni_dp = 1E17\n")
        f.write("#nu_ni = 3.0\n")
        f.write("#ne0 = 3.629E19\n")
        f.write("#ne9 = 1.523E19\n")
        f.write("#ne_sep = 0.3E19\n")
        f.write("#ne_dp = 1E17\n")
        f.write("#nu_ne = 2.5\n")
        f.write("#Ti0 = 35\n")
        f.write("#Ti9 = 6\n")
        f.write("#Ti_sep = 0.6\n")
        f.write("#Ti_dp = 0.03\n")
        f.write("#nu_Ti = 3.5\n")
        f.write("#Te0 = 36\n")
        f.write("#Te9 = 6\n")
        f.write("#Te_sep = 0.6\n")
        f.write("#Te_dp = 0.01\n")
        f.write("#nu_Te = 3.5 \n")
        f.write("\n")

        f.write("#CURRENT-RELATED PARAMETERS \n")
        f.write("IP = %s \n" % str(data.plasmaCur))
        f.write("\n")
        f.write("#ION ORBIT LOSS CALCULATION \n")
        f.write("numcos = 8 \n")
        f.write("R_loss = 0.5 \n")
        f.write("\n")

        f.write("pfr_ni_val = 1.0E14 \n")
        f.write("pfr_ne_val = 1.0E14 \n")
        f.write("pfr_Ti_val = 0.002  \n")
        f.write("pfr_Te_val = 0.002  \n")

        f.write("############################################################################### \n")
        f.write("# CONFIG - YOU CAN PROBABLY LEAVE THESE ALONE IF YOU DON'T KNOW WHAT THEY ARE \n")
        f.write("############################################################################### \n")
        f.write("verbose = 1 \n")

        f.write("sollines_psi_max = 1.07 \n")
        f.write("num_sollines = 6 \n")

        f.write("xi_ib_pts = 10 \n")
        f.write("xi_ob_pts = 10 \n")

        f.write("core_thetapts_ntrl = 50 \n")

        f.write("#rhopts_ntrl = 100 \n")
        f.write("edge_rho_ntrl = 0.8 \n")
        f.write("rhopts_edge_ntrl = 5 \n")
        f.write("rhopts_core_ntrl = 10 \n")
        f.write("ib_div_pol_pts = 7 \n")
        f.write("ob_div_pol_pts = 7 \n")
        f.write("wall_ni_min = 1.0E15 \n")
        f.write("wall_ne_min = 1.0E15 \n")
        f.write("wall_Ti_min = %s \n" % str(0.02 * 1.0E3 * 1.6021E-19))
        f.write("wall_Te_min = %s \n" % str(0.02 * 1.0E3 * 1.6021E-19))
        f.write("core_thetapts_ntrl = 30 \n")

        f.close()


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from contours.quad import QuadContourGenerator

def draw_contour_line(R, Z, array, val, pathnum):
    c = QuadContourGenerator.from_rectilinear(R[0], Z[:, 0], array)

    res = c.contour(val)[pathnum]
    x = res[:, 0]
    y = res[:, 1]
    return x, y


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from math import degrees, sqrt, acos

def getangle3ptsdeg(p1, p2, p3):
    a = sqrt((p1[0]-p2[0])**2+(p1[1]-p2[1])**2)
    b = sqrt((p2[0]-p3[0])**2+(p2[1]-p3[1])**2)
    c = sqrt((p1[0]-p3[0])**2+(p1[1]-p3[1])**2)
    theta = degrees(acos((c**2 - a**2 - b**2)/(-2*a*b)))  # returns degree in radians
    return theta



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np

def calc_pwr_frac(ebeam):
    # calculate pwr fracs. These fits are only useful for DIII-D

    pwr_frac = np.zeros(3)
    pwr_frac[0] = (68 + 0.11 * ebeam) / 100
    pwr_frac[1] = (-159 + 6.53 * ebeam - 0.082 * (ebeam ** 2) + 0.00034 * (ebeam ** 3)) / 100
    pwr_frac[2] = (191 - 6.64 * ebeam + 0.082 * (ebeam ** 2) - 0.00034 * (ebeam ** 3)) / 100

    return pwr_frac



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
import os


def create_triangle_infile(all_pts_unique, all_segs_unique, core):
    # OUTPUT .poly FILE AND RUN TRIANGLE PROGRAM
    open('./outputs/exp_mesh.poly', 'w').close()
    outfile = open('./outputs/exp_mesh.poly', 'ab')
    filepath = os.path.realpath(outfile.name)
    np.savetxt(outfile,
               np.array([all_pts_unique.shape[0], 2, 0, 0])[None],
               fmt='%i %i %i %i')
    np.savetxt(outfile,
               np.column_stack((np.arange(len(all_pts_unique)),
                                all_pts_unique)),
               fmt='%i %f %f')
    np.savetxt(outfile,
               np.array([all_segs_unique.shape[0], 0])[None],
               fmt='%i %i')
    np.savetxt(outfile,
               np.column_stack((np.arange(len(all_segs_unique)),
                                all_segs_unique,
                                np.zeros(len(all_segs_unique), dtype='int'))),
               fmt='%i %i %i %i')
    np.savetxt(outfile,
               np.array([1])[None],
               fmt='%i')
    np.savetxt(outfile,
               np.array([1, core.pts.axis.mag[0], core.pts.axis.mag[1]])[None],
               fmt='%i %f %f')
    np.savetxt(outfile,
               np.array([0])[None],
               fmt='%i')
    outfile.close()

    return filepath



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Sat Mar 24 13:25:51 2018

@author: max
"""
from __future__ import division
from subprocess import call
from collections import namedtuple
import numpy as np
from scipy.interpolate import interp1d, interp2d, UnivariateSpline, LinearNDInterpolator
import os
import re
import sys
import pickle


class ImpRad:
    """Calculates the impurity charge state densities and resulting impurity emissivities.

    Attributes:
        Lz
        dLzdT

    Methods:
        __init__
        calc_Lz
        frac_abun
    """

    def __init__(self, z=None, core=None):
        sys.dont_write_bytecode = True

        # list of impurity names and pickled impurity interpolation object filenames, if they exist. Will be used later.
        imp_names = {}
        imp_names[2] = 'Helium'
        imp_names[4] = 'Beryllium'
        imp_names[6] = 'Carbon'
        imp_names[8] = 'Oxygen'
        imp_names[10] = 'Neon'
        imp_names[18] = 'Argon'
        imp_names[36] = 'Krypton'
        imp_names[54] = 'Xenon'
        imp_names[74] = 'Tungsten'

        if (core is not None) and (z is not None):
            print """Both core and z were specified. z will be ignored and core Lz parameters will be updated.'
                  'If you would like a specific z only, then instantiate ImpRad as a standalone instance and'
                  'pass z, but not a core instance."""
        if core is not None:
            for z in [6, 4, 74, 10, 18, 36]:  # list of all z elements in core. Update this list as necessary.
                try:  # before running adpak, try to find a pickled interpolator somewhere in the main directory.
                    Lz, dLzdT = self.find_interp(z, imp_names)
                except:
                    print 'Pickled interpolater not found for {}. Running adpak.'.format(imp_names[z])
                    Lz, dLzdT = self.run_adpak(z, imp_names)
                core.update_Lz_data(z, Lz, dLzdT)

        elif z is not None:
            try:  # before running adpak, try to find a pickled interpolator somewhere in the main directory.
                print 'trying to find interpolation object'
                self.Lz, self.dLzdT = self.find_interp(z, imp_names)
            except:
                print 'didn\'t work'
                print 'Pickled interpolater not found for {}. Running adpak.'.format(imp_names[z])
                self.Lz, self.dLzdT = self.run_adpak(z, imp_names)
        else:
            raise Exception("Neither core nor z were specified. I can\'t read minds. Stopping.")

        # TODO: Convert adpak routines to python binary using f2py so we can eliminate the main.f driver program

    def find_interp(self, z, imp_names):
        # create filename
        Lz_pkl_file = imp_names[z] + '_Lz.pkl'
        dLzdT_pkl_file = imp_names[z] + '_dLzdT.pkl'

        # search for pkl_file
        outfile_found = 0
        for root, subdirs, files in os.walk(os.getcwd()):
            for filename in files:
                if filename == Lz_pkl_file:
                    Lz_found = True
                    os.path.join(root, filename)
                    pkl_file_loc = os.path.join(root, filename)
                    pickle_in = open(pkl_file_loc, "rb")
                    Lz_interp = pickle.load(pickle_in)
                    pickle_in.close()
                if filename == dLzdT_pkl_file:
                    dLzdT_found = True
                    os.path.join(root, filename)
                    pkl_file_loc = os.path.join(root, filename)
                    pickle_in = open(pkl_file_loc, "rb")
                    dLzdT_interp = pickle.load(pickle_in)
                    pickle_in.close()
        if Lz_found and dLzdT_found:
            return Lz_interp, dLzdT_interp
        else:
            # pickle file was not found. Bummer.
            raise ValueError('One or more of the pickle files was not found for {}.'.format(imp_names[z]))

    def run_adpak(self, z, imp_names):

        # specify input parameters. These could easily be moved into an input file if necessary.
        inp = {}
        inp['z_imp'] = z  # impurity z
        inp[
            'laden'] = laden  # Flag for excitation energy and ion. potential calc. If 0 --> use Mayer formalism. If 1 --> use More formalism
        inp['ladtip'] = ladtip  # If 1 --> use tabulated ionization potentials
        inp[
            'leci'] = leci  # Flag for calculation of ionization rates: 1: XSNQ, 2: Belfast group (for H through O) 3: Younger (Scandium and Fe)
        inp[
            'ldrmlt'] = ldrmlt  # Dielectronic multiplier. 0: Use CDNN, CDNM arrays as given# 1 --> Set CDNN, CDNM equal to 1, 2 --> Use Y. Hahn factors (dut) to set up CDNN, CDNM
        inp['ncxopt'] = ncxopt  # Selects cross sections to be used   1 --> OSAS   2 --> GJ    3 --> OSCT
        inp['imode'] = 1
        inp['nmin_T'] = -3
        inp['nmax_T'] = 2
        inp['nmin_n'] = 13
        inp['nmax_n'] = 15
        inp['nte'] = 25  # Number of electron temperatures in table
        inp['nne'] = 2  # Number of electron densities in table
        inp['tei'] = np.logspace(inp['nmin_T'], inp['nmax_T'], inp['nte'])  # Array of nte Te values in keV
        inp['anei'] = np.logspace(inp['nmin_n'], inp['nmax_n'], inp['nne'])  # Array of nne ne values in cm^-3
        inp['ncxb'] = 1  # Number of NB components (if 0 --> No NB's_)
        inp['ivunit'] = 2  # Units for NB energy  1 --> cm/s  2 --> keV / amu

        # Some notes about the adpak interface:
        #   - Lz is a function of Te, nn/ne, and Tn
        #   - As a result, you don't need to run multiple ne values, however 2 are necessary for adpak to run smoothly
        #     for reasons that aren't totally clear. We will use ne=1E13 cm^-3 (1E19 m^3)
        #   - To get the values necessary to do a 3-D interpolation, we'll loop over nn/ne and Tn (in kev/amu)
        #   - Lz(nn/ne) doesn't change much for nn/ne > 1, which is a situation that only typically occurs in the PFR
        #       - we will loop over ~10 nn/ne ratios ranging from 0 to 1, i.e. (10^-6, 10^-5, ...etc.)
        #   - for a given nn/ne ratio, increasing Tn tends to INCREASE the effects of the neutrals
        #       - increasing Tn beyond ~100kev/amu has little effect on Lz, even for small nn/ne.
        #       - we will loop over  ~10 Tn values ranging from 0.001 to 100 kev/amu (i.e. 0.001, 0.01, 0.1, ...)

        Tn_num = 8
        nf_num = 8
        Te_num = 100

        Tn_vals = np.logspace(-3, 2, Tn_num)
        nf_vals = np.logspace(-7, 0, nf_num)
        Te_vals = np.logspace(-3, 2, Te_num)

        Lz_complete = np.zeros((Tn_num, nf_num, Te_num))
        dLzdT_complete = np.zeros((Tn_num, nf_num, Te_num))

        for i, nf in enumerate(nf_vals):
            for j, Tn in enumerate(Tn_vals):

                inp['anneut'] = nf * 1 * 10 ** 13  # inp['nmin_n']
                inp['vneut'] = Tn

                inp2 = namedtuple('inp2', inp.keys())(*inp.values())

                # prepare input file
                f = open('./toadpak', 'w')
                f.write(' &inp')
                f.write('\n' + '  inucz = ' + str(inp2.z_imp))
                f.write('\n' + '  zte = ' + str(0))  # this parameter doesn't matter because imode = 1
                f.write('\n' + '  zne = ' + str(0))  # this parameter doesn't matter because imode = 1
                f.write('\n' + '  laden = ' + str(inp2.laden))
                f.write('\n' + '  ladtip = ' + str(inp2.ladtip))
                f.write('\n' + '  leci = ' + str(inp2.leci))
                f.write('\n' + '  ldrmlt = ' + str(inp2.ldrmlt))
                f.write('\n' + '  ncxb = ' + str(inp2.ncxb))
                f.write('\n' + '  ncxopt = ' + str(inp2.ncxopt))
                f.write('\n' + '  ivunit = ' + str(inp2.ivunit))
                f.write('\n' + '  anneut = ' + str(inp2.anneut))
                f.write('\n' + '  vneut = ' + str(inp2.vneut))
                f.write('\n' + '  imode = ' + str(inp2.imode))
                f.write('\n' + '  nte = ' + str(inp2.nte))
                f.write('\n' + '  nne = ' + str(inp2.nne))
                # f.write('\n' + '  tei = ' + ' '.join(map(str, inp2.tei)))
                # f.write('\n' + '  tei = ' +  np.array2string(inp2.tei, precision=3, separator=' ',suppress_small = False))
                print np.array2string(inp2.tei, formatter={'float_kind': lambda x: "%.5f" % x})[1:-1]
                f.write(
                    '\n' + '  tei = ' + np.array2string(inp2.tei, formatter={'float_kind': lambda x: "%.5f" % x})[1:-1])
                f.write('\n' + '  anei = ' + ' '.join(map(str, inp2.anei)))
                f.write('\n' + '  nmin = ' + str(inp2.nmin_T))
                f.write('\n' + '  nmax = ' + str(inp2.nmax_T))
                f.write('\n' + ' $end')
                f.write('\n' + '')
                f.write('\n')
                f.close()
                # sys.exit()
                # call adpak
                try:
                    call([inp2.adpak_loc + 'adpak', os.getcwd() + '/toadpak'])
                except AttributeError:
                    try:
                        call(['adpak', os.getcwd() + '/toadpak'])
                    except:
                        raise RuntimeError("Unable to locate adpak.")

                # read adpak output
                with open(os.getcwd() + '/outblk.dat', 'r') as f:
                    # read entire file into one long string variable
                    data = f.read().replace('\n', ' ').replace('. ', ' ').replace(', ', ' ')

                # the following arrays all have the same dimension despite the fact that alinzr
                # is a function only of Te
                # array[ charge state, density value, temperature value ]
                alinzr = np.zeros((inp2.z_imp + 1, inp2.nne, inp2.nte))
                alradr = np.zeros((inp2.z_imp + 1, inp2.nne, inp2.nte))
                alrecr = np.zeros((inp2.z_imp + 1, inp2.nne, inp2.nte))

                # find altei data
                result = re.match(r'.*?data \(+%s.*?\/(.*?)\/.*' % 'altei', data).group(1)
                altei = np.asarray(result.split(), dtype=float)

                # find alnei data
                result = re.match(r'.*?data \(+%s.*?\/(.*?)\/.*' % 'alnei', data).group(1)
                alnei = np.asarray(result.split(), dtype=float)

                # for each charge state plus the ground state,
                for cs in np.arange(inp2.z_imp + 1):
                    # find alinzr data for cs. alinzr is a function of temperature only
                    result = re.match(r'.*?data \(+%s\(\d+ +%s.*?\/(.*?)\/.*' % ('alinzr', cs + 1), data).group(1)
                    alinzr[cs, :, :] = np.asarray(result.split(), dtype=float).reshape(-1, inp2.nte)

                    # find alradr data for cs. alradr is a function of both temperature and electron density
                    result = re.match(r'.*?data \(+%s\(\d+ +%s.*?\/(.*?)\/.*' % ('alradr', cs + 1), data).group(1)
                    alradr[cs, :, :] = np.asarray(result.split(), dtype=float).reshape(-1, inp2.nte)

                    # find alrecr data for cs. alrecr is a function of both temperature and electron density
                    result = re.match(r'.*?data \(+%s\(\d+ +%s.*?\/(.*?)\/.*' % ('alrecr', cs + 1), data).group(1)
                    alrecr[cs, :, :] = np.asarray(result.split(), dtype=float).reshape(-1, inp2.nte)

                # assemble data to pass to Lz and charge state density calculations
                data = namedtuple("data", 'altei alnei alinzr alradr alrecr')(altei, alnei, alinzr, alradr, alrecr)

                # calculate impurity charge state densities. Uncomment the following line if desired.
                # self.calc_nz_cs = self.nz_cs(inp, data)

                # create interpolation function for Lz and dLzdT based on coronal equilibrium approximation
                Lz_interp, dLzdT_interp = self.calc_Lz(inp2, data)

                Lz_complete[i, j, :] = Lz_interp(Te_vals * 1E3 * 1.6021E-19)
                dLzdT_complete[i, j, :] = dLzdT_interp(Te_vals * 1E3 * 1.6021E-19)

        Tn_mesh, nf_mesh, Te_mesh = np.meshgrid(Tn_vals, nf_vals, Te_vals)

        points = np.column_stack((np.log10(Tn_mesh).flatten(),
                                  np.log10(nf_mesh).flatten(),
                                  np.log10(Te_mesh).flatten()))
        Lz = LinearNDInterpolator(points, Lz_complete.flatten(), rescale=False)
        dLzdT = LinearNDInterpolator(points, dLzdT_complete.flatten(), rescale=False)

        # pickle the interpolation object to save time in the future
        filename = os.getcwd() + '/Lz_interpolators/' + imp_names[z] + '_Lz.pkl'
        outfile = open(filename, 'wb')
        pickle.dump(Lz, outfile)
        outfile.close()

        filename = os.getcwd() + '/Lz_interpolators/' + imp_names[z] + '_dLzdT.pkl'
        outfile = open(filename, 'wb')
        pickle.dump(dLzdT, outfile)
        outfile.close()

        # cleanup
        files = ['toadpak', 'outplt.txt', 'outblk.dat', 'adfits.txt', 'outadpk.txt']
        for f in files:
            try:
                os.remove(f)
            except OSError:
                pass

        return Lz, dLzdT

    @classmethod
    def calc_nz_cs(cls, inp, data):
        """calculates impurity densities by charge state"""

        tei_lin = np.linspace(inp.nmin, inp.nmax, inp.nte)

        nz_cs = []

        # for each charge state,
        for cs in np.arange(inp.z_imp + 1):
            # get fractional abundances of each as a function of ne and Te
            frac_abun_interp = interp2d(tei_lin,
                                        inp.anei,
                                        cls.frac_abun(inp, data)[:, :, cs])

            # get the fractional abundance for each value in the nz array
            nz_frac_abun = np.zeros(brnd.nz.shape)
            for (i, j), brnd.nz in np.ndenumerate(brnd.nz):
                nz_frac_abun[i, j] = frac_abun_interp(np.log10(brnd.Te_kev[i, j]),
                                                      brnd.ne[i, j])

            # multiply the fractional abundance by the overall impurity density to get the
            # density of this charge state
            nz_cs.append(nz_frac_abun * brnd.nz)

        return nz_cs

    @classmethod
    def calc_Lz(cls, inp, data):
        """creates the emissivity function Lz over a range of temperatures"""

        def kev_J(x):
            return x * 1E3 * 1.6021E-19

        tei_lin = np.linspace(inp.nmin_T, inp.nmax_T, inp.nte)
        Lz_tot = np.zeros(inp.nte)

        # for each charge state,
        for cs in np.arange(inp.z_imp + 1):
            # get fractional abundances of each as a function of ne and Te
            frac_abun_interp = interp1d(tei_lin,
                                        cls.frac_abun(inp, data)[0, :, cs])

            Lz_cs_interp = interp1d(tei_lin,
                                    data.alradr[cs, 0, :])

            for i, T in enumerate(tei_lin):
                Lz_tot[i] = Lz_tot[i] + 10 ** Lz_cs_interp(T) * frac_abun_interp(T)

        # spline fit the logarithm of the emissivity. We'll spline fit the actual values next
        Lz_tot_interp = UnivariateSpline(inp.tei, np.log10(Lz_tot) + 100.0, s=0)

        # the above function gives the right values, but we also need the derivatives
        # we will now do a spline fit of the values themselves, rather than their base-10 logarithm
        T_kev_vals = np.logspace(-3, 2, 10000)
        T_J_vals = kev_J(T_kev_vals)
        Lz_vals = 10.0 ** (Lz_tot_interp(T_kev_vals) - 100.0)

        Lz_interp = UnivariateSpline(T_J_vals, Lz_vals, s=0)
        dLzdT_interp = Lz_interp.derivative()

        return Lz_interp, dLzdT_interp

    @staticmethod
    def frac_abun(inp, data):
        """Calculates the fractional abundances of impurity charge states using a coronal equilibrium approximation."""

        def null(M, eps=1e-20):
            u, s, vh = np.linalg.svd(M)
            return vh[np.argmin(np.abs(s))]

        # Create nchrgsr x nchrgsr array
        M = np.zeros((inp.z_imp + 1, inp.z_imp + 1))

        frac_abun = np.zeros((inp.nne, inp.nte, inp.z_imp + 1))

        # for i, dens in enumerate(inp.anei):
        #     for j, temp in enumerate(inp.tei):
        #         for (cs, col), value in np.ndenumerate(M):
        #             if cs == 0:  # do the first row differently
        #                 M[cs, cs] = -(10**data.alinzr[cs, i, j] + 10**data.alrecr[cs, i, j])
        #                 M[cs, cs + 1] = 10 ** data.alrecr[cs + 1, i, j]
        #             elif cs == inp.z_imp:  # do the last row differently
        #                 M[cs, cs - 1] = 10 ** data.alinzr[cs - 1, i, j]
        #                 M[cs, cs] = -(10 ** data.alinzr[cs, i, j] + 10 ** data.alrecr[cs, i, j])
        #             else:
        #                 M[cs, cs - 1] = 10 ** data.alinzr[cs - 1, i, j]
        #                 M[cs, cs] = -(10 ** data.alinzr[cs, i, j] + 10 ** data.alrecr[cs, i, j])
        #                 M[cs, cs + 1] = 10 ** data.alrecr[cs + 1, i, j]

        for i, dens in enumerate(inp.anei):
            for j, temp in enumerate(inp.tei):

                for (cs, col), value in np.ndenumerate(M):
                    if cs == 0:  # do the first row differently
                        M[cs, cs] = -(10 ** data.alinzr[cs, i, j])
                        M[cs, cs + 1] = 10 ** data.alrecr[cs + 1, i, j]
                    elif cs == inp.z_imp:  # do the last row differently
                        M[cs, cs - 1] = 10 ** data.alinzr[cs - 1, i, j]
                        M[cs, cs] = -(10 ** data.alrecr[cs, i, j])
                    else:
                        M[cs, cs - 1] = 10 ** data.alinzr[cs - 1, i, j]
                        M[cs, cs] = -(10 ** data.alinzr[cs, i, j] + 10 ** data.alrecr[cs, i, j])
                        M[cs, cs + 1] = 10 ** data.alrecr[cs + 1, i, j]
                # the following line may help convergence. Uncomment it if necessary. Everything
                # gets normalized anyway.
                # M = M*1E20

                # print M
                # sys.exit()

                result = np.abs(null(M).T)  # solve(M, b)
                # print 'result = ',result
                result = np.where(result < 0, -result, result)
                frac_abun[i, j, :] = result / np.sum(result)

        return frac_abun


if __name__ == '__main__':
    """Example script."""

    # input parameters. Note that some parameters are hard coded in ImpRad.__init__
    z_imp = 18  # Atomic number, Z, of desired element
    laden = 1  # Flag for excitation energy and ion. potential calc. If 0 --> use Mayer formalism. If 1 --> use More formalism
    ladtip = 1  # If 1 --> use tabulated ionization potentials
    leci = 1  # Flag for calculation of ionization rates: 1: XSNQ, 2: Belfast group (for H through O) 3: Younger (Scandium and Fe)
    ldrmlt = 2  # Dielectronic multiplier. 0: Use CDNN, CDNM arrays as given# 1 --> Set CDNN, CDNM equal to 1, 2 --> Use Y. Hahn factors (dut) to set up CDNN, CDNM
    ncxopt = 1  # Selects cross sections to be used   1 --> OSAS   2 --> GJ    3 --> OSCT

    # Helium
    He_2 = ImpRad(z=2)

    # Beryllium
    Be_4 = ImpRad(z=4)

    # Carbon
    C_6 = ImpRad(z=6)

    # Oxygen
    O_8 = ImpRad(z=8)

    # #Neon
    Ne_10 = ImpRad(z=10)

    # #Argon
    Ar_18 = ImpRad(z=18)

    # #Krypton
    Kr_36 = ImpRad(z=36)

    # Xenon
    Xe_54 = ImpRad(z=54)

    # Tungsten
    W_74 = ImpRad(z=74)


    def element_plot(inst, element):
        # specify density and temperature parameters at which you want to
        # evaluate Lz, dLzdT, etc. and convert to dict and namedtuple
        Te_kev = np.logspace(-3, 2, 100)
        Tn = np.full(Te_kev.shape, 1E-3)

        # prepare figure
        # fig = plt.figure(figsize=(6, 4))
        # ax1 = fig.add_subplot(1, 1, 1)
        # ax1.set_title('{} $L_z$ as a Function of Temperature'.format(element))
        # ax1.set_xlabel(r'Electron Temperature ($keV$)')
        # ax1.set_ylabel(r'$L_z$ ($W*m^3$)')
        #
        # nf = np.full(Tn.shape, 1E-7)
        # ax1.loglog(Te_kev, inst.Lz(np.log10(Tn), np.log10(nf), np.log10(Te_kev)))

        # for i,v in enumerate(np.logspace(-5,-1,5)):
        # nf = np.full(Tn.shape, v)
        # if i==0:
        # ax1.loglog(Te_kev, inst.Lz(np.log10(0.002), np.log10(nf), np.log10(Te_kev)))
        # clean up and show plot
        # plt.tight_layout()
        # fig.savefig('/home/max/Documents/{}_Lz.png'.format(element))
        # plt.show()
        # return Te_kev, inst.Lz(np.log10(0.002), np.log10(nf), np.log10(Te_kev))

    # element_plot(He_2,'Helium')
    # element_plot(C_6,'Carbon')
    # element_plot(Ne_10,'Neon')
    # element_plot(Ar_18,'Argon')
    # element_plot(Kr_36,'Krypton')
    # element_plot(Xe_54,'Xenon')
    # element_plot(W_74,'Tungsten')



#/usr/bin/python

from matplotlib import pyplot as plt
from numpy import genfromtxt, savetxt
import numpy as np
import os
from scipy.interpolate import LSQUnivariateSpline

class dataSmoother:

    def __init__(self, file, knots=[]):
        self.filename = file
        with open(file, "r") as f:
            inData = genfromtxt(f, float, comments='#').T
            self.rho = inData[0]
            self.orig_data = inData[1]

        self.spline = LSQUnivariateSpline(self.rho, self.orig_data, knots, k=5)

    def plot_orig(self):
        fig = self._plot_base()
        fig.scatter(self.rho, self.orig_data, color='red')
        plt.show()
        return fig

    def plot(self):
        fig = self._plot_base()

        fig.scatter(self.rho, self.orig_data, color='red')
        fig.plot(self.rho, self.spline(self.rho), color='black')
        fig.scatter(self.rho, self.spline(self.rho), color='green', marker="x")

        fig.legend(["Spline", "Original", "Cleaned"])

        plt.show()
        return fig

    def save(self):
        answer = ""
        while answer not in ["y", "n"]:
            answer = raw_input("OK to continue [Y/N]? ").lower()
        print "Backing up old file"
        os.rename(self.filename, self.filename + '.bak')
        contents = np.array([self.rho, self.spline(self.rho)]).T
        savetxt(self.filename + "_clean", contents, delimiter='    ')
    def get_spline(self):
        return self.spline

    def set_knots(self, knots):
        self.spline = LSQUnivariateSpline(self.rho, self.orig_data, knots, k=5)
        print "knots set"

    def _plot_base(self):
        plot = plt.figure()
        fig = plot.add_subplot(111)
        fig.set_xlabel(r'$\rho', fontsize=30)
        fig.set_ylabel(r'Data', fontsize=30)
        plt.xticks(fontsize=20)
        plt.yticks(fontsize=20)

        return fig

    def plot_derivative(self):
        fig = self._plot_base()
        fig.scatter(self.rho, self.spline.derivative()(self.rho), color="blue")
        return fig



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from math import pi, ceil, atan2


def calc_theta1d(pts, thetapts_approx):
    # define theta points
    def atan3(y, x):
        result = atan2(y, x)
        if result < 0:
            result = result + 2 * pi
        return result

    if pts.xpt[1] is None or pts.xpt[0] is None:
        # Grab the actual x-point
        if pts.xpt[0] is not None:
            xpt_temp = pts.xpt[0]
        else:
            xpt_temp = pts.xpt[1]
    else:
        # THere are 2 xpoints. Use the bottom one.
        xpt_temp = pts.xpt[0]

    if xpt_temp[1] > 0.:
    # these theta markers correspond to the obmp, xpt, ibmp, bot, and obmp+2pi, respectively
        theta_markers = np.zeros(5)
        theta_markers[0] = atan2((pts.obmp[1] - pts.axis.geo[1]), (pts.obmp[0] - pts.axis.geo[0]))
        theta_markers[1] = atan3((xpt_temp[1] - pts.axis.geo[1]), (xpt_temp[0] - pts.axis.geo[0]))
        theta_markers[2] = atan3((pts.ibmp[1] - pts.axis.geo[1]), (pts.ibmp[0] - pts.axis.geo[0]))
        theta_markers[3] = atan3((pts.bottom[1] - pts.axis.geo[1]), (pts.bottom[0] - pts.axis.geo[0]))
        theta_markers[4] = theta_markers[0] + 2 * pi

    else:
        # these theta markers correspond to the obmp, top, ibmp, xpt, and obmp+2pi, respectively
        theta_markers = np.zeros(5)
        theta_markers[0] = atan2((pts.obmp[1] - pts.axis.geo[1]), (pts.obmp[0] - pts.axis.geo[0]))
        theta_markers[1] = atan3((pts.top[1] - pts.axis.geo[1]), (pts.top[0] - pts.axis.geo[0]))
        theta_markers[2] = atan3((pts.ibmp[1] - pts.axis.geo[1]), (pts.ibmp[0] - pts.axis.geo[0]))
        theta_markers[3] = atan3((xpt_temp[1] - pts.axis.geo[1]), (xpt_temp[0] - pts.axis.geo[0]))
        theta_markers[4] = theta_markers[0] + 2 * pi

    try:
        min_delta_theta = 2 * pi / thetapts_approx
    except:
        print 'thetapts_approx not defined. Setting to 30'
        min_delta_theta = 2 * pi / 30

    theta1d = np.zeros(0)
    for i in range(4):
        quad_pts = int(ceil((theta_markers[i + 1] - theta_markers[i]) / min_delta_theta))
        if i == 3:
            quad_theta = np.linspace(theta_markers[i], theta_markers[i + 1], quad_pts+1, endpoint=True)
        else:
            quad_theta = np.linspace(theta_markers[i], theta_markers[i + 1], quad_pts, endpoint=False)
        theta1d = np.concatenate((theta1d, quad_theta))
    return theta1d, theta_markers


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from collections import namedtuple
from GT3.Core.Functions.CalcFSA import calc_fsa
import numpy as np

def corePatch(core, neutFlag=True):
    """
    Updates deuterium ion density and zeff. D density is invalid if no D density file is given because it will be
    set to 0. This screws up subsequent calculations in non-obvious ways. As an example, z_eff calculation will
    include D density as 0s but still have a carbon density from the fracz input file, giving an invalid result

    This is addressed by recreating the n_fsa namedtuple and z_eff_fsa in full, as namedtuples cannot be changed piecemeal

    :param neutFlag:
    :param core:
    :return:
    """


    if neutFlag:
        core.n = namedtuple('n', 'i e n C')(core.n.e/(1.+.025*6.0), core.n.e, core.n.n, 0.025 * core.n.e/(1.+.025*6.0))   # TODO: Update 0.025 and 6.0 to actual fracz and zbar2
    else:
        core.n = namedtuple('n', 'i e n C')(core.n.e / (1. + .025 * 6.0), core.n.e,
                                   namedtuple('n', 's t tot')(
                                       np.zeros(core.n.i.shape),  # slow
                                       np.zeros(core.n.i.shape),  # thermal
                                       np.zeros(core.n.i.shape)),  # total
                                   0.025 * core.n.e / (1. + .025 * 6.0))
    core.n_fsa = namedtuple('n', 'i e n C')(
        calc_fsa(core.n.i, core.R, core.Z),
        calc_fsa(core.n.e, core.R, core.Z),
        namedtuple('n', 's t tot')(
            calc_fsa(core.n.n.s, core.R, core.Z),  # slow
            calc_fsa(core.n.n.t, core.R, core.Z),  # thermal
            calc_fsa(core.n.n.tot, core.R, core.Z)  # total
        ),
        calc_fsa(core.n.C, core.R, core.Z))

    core.z_eff_fsa = calc_fsa((core.n.i * (1.**2) + core.n.C * (6.0**2))/(core.n.i * 1.0 + core.n.C * 6.0), core.R, core.Z) #TODO Similar updates (1.0 = atnum, 6.0 = zbar2)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from GT3.Functions.GetNum import getNum
from collections import namedtuple

def getVals(s, t, f):
    """

    This function interactively obtains the 0D parameters for your input file to be created. The commented-out section
    is data for a shot just for debugging purposes

    :param s: shot id
    :param t: time id
    :param f: file to be generated
    :return: Your mom
    """

    data ={}
    """ Function that gets 0D plasma values from user input """

    print """Enter 0D plasma parameters for shot %s.%s

             File will be generated at inputs/%s""" % (str(s), str(t), str(f))

    # data={'eq1' : 1.6E-19,
    #       'eq2' : 9.6E-19,
    #       'xmas1' : 3.35E-27,
    #       'xmas2' : 2.01E-26,
    #       'xk' : 1.6E-19,
    #       'delma' : 0.005,
    #       'aminor' : 0.598,
    #       'bphi' : -2.04,
    #       'rmajor' : 1.7,
    #       'kappaup': 1.82,
    #       'kappalo' : 1.82,
    #       'triup': .237,
    #       'trilo' : 0.,
    #       'thetapts' : 30,
    #       'rhopts_edge' : 100,
    #       'rhopts_core' : 10,
    #       'ebeam' : 77.48,
    #       'pbeam' : 4.6,
    #       'rtang' : 1.09,
    #       'bknot' : 2.0,
    #       'pwrfrac1' : .76,
    #       'pwrfrac2' : .13,
    #       'pwrfrac3' : .11,
    #       'epsknot' : 1.265,
    #       'epssep' : 1.82,
    #       'shftknot' : 0.033
    #       }

    data['xk'] = 1.6E-19
    data['delma'] = .005
    data['aminor'] = getNum("Enter the plasma radius (a minor): " ,"f")
    data['bphi'] = getNum("Enter the toroidal plasma field strength in T (abs mag): ", "f")
    data['rmajor'] = getNum("Enter R major: ", "f")
    data['kappaup'] = getNum("Enter upper elongation (kappa): ", "f")
    data['kappalo'] = getNum("Enter lower elongation (kappa): ", "f")
    data['triup'] = getNum("Enter upper triangularity (delta): ", "f")
    data['trilo'] = getNum("Enter lower triangularity (delta): ", "f")
    data['thetapts'] = getNum("Enter approximate number of theta points (typically 30): ", 'i')
    data['rhopts_edge'] = getNum("Enter rho points in the edge (typically 100): ", 'i')
    data['rhopts_core'] = getNum("Enter rho points in the core (typically 10): ", 'i')
    data['xptR'] = getNum("Enter the X-point R coordinate: ", 'f')
    data['xptZ'] = getNum("Enter the X-point Z coordinate: ", 'f')
    data['jknot'] = getNum("Enter the r=0 plasma current density (in A/m^3): ", 'f')
    data['plasmaCur'] = getNum("Enter the plasma current (in MA): ", 'f')
    data['ebeam'] = getNum("Enter beam ion energy in eV: ", 'f')
    data['pbeam'] = getNum("Enter beam power in MWYup: ", 'f')
    data['rtang'] = getNum("Enter radius of tangency in cm: ", 'f')
    data['bknot'] = abs(data['bphi'])
    data['pwrfrac1'] = getNum("Enter fraction of beam power to D1: ", 'f')
    data['pwrfrac2'] = getNum("Enter fraction of beam power to D2: ", 'f')
    data['pwrfrac3'] = getNum("Enter fraction of beam power to D3: ", 'f')
    data['epsknot'] = getNum("Enter epsilon at the plasma center: ", 'f')
    data['epssep'] = getNum("Enter epsilon at separatrix: ", 'f')
    data['shftknot'] = getNum("Enter the Shavranof shift at the plasma center: ", 'f')

    return namedtuple('data', sorted(data))(**data)


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import griddata

def grid(x, y, z, resX=100, resY=100):
    """Convert 3 column data to matplotlib grid"""
    xi = np.linspace(min(x), max(x), resX)
    yi = np.linspace(min(y), max(y), resY)
    X, Y = np.meshgrid(xi, yi)
    Z = griddata(np.column_stack((x, y)), z, (X, Y), method='linear')

    #interp = Rbf(x, y, z, function='linear')
    #Z = interp(X, Y)
    return X, Y, Z


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
import numpy as np

# Calculates the radial electric field including the J x B force from IOL

def calc_Er_iol(n_i, n_e, m_i, n_n, Bphi, Lp_i, dp_dr, e_i, T_i, dn_dr, izn_rate, Jr_iol):
    Jr_visc = 0.0
    Jr_neut = -Jr_iol - Jr_visc
    iol_term = (Jr_neut * Bphi**2.0) / (m_i * izn_rate)  # Both the ion and neutral densities are built into izn_rate
    diamag_term = -1.0 * (Lp_i * T_i.J) / e_i  # Pressure gradient is negative, so multiply by -1
    diamag_term_orig = -1.0 * (dp_dr / (e_i * n_i))  # Original form of diamagnetic term
    try:
        dn_dr
        neut_dens_term = -1.0 * (T_i.J * dn_dr) / (e_i * n_n.t)
    except:
        neut_dens_term = [0.] * len(n_i)
    E_r_iol = iol_term + diamag_term + neut_dens_term

    return E_r_iol, iol_term, diamag_term, diamag_term_orig, neut_dens_term



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from shapely.geometry import Point, LineString, LinearRing
from GT3.Core.Functions.Cut import cut
from collections import namedtuple
import matplotlib.pyplot as plt
from matplotlib.path import Path


def calc_pts_lines(psi_data, xpt, wall, mag_axis, sep_val, core=None):

    """
    Calculates points and lines for the flux surfaces

    :param psi_data:
    :type psi_data: psi
    :param xpt:
    :param wall:
    :type wall: LineString
    :param mag_axis:
    :return:
    :rtype: object
    """
    Points = namedtuple('Points', 'ibmp obmp top bottom xpt axis strike')
    PlasmaAxis = namedtuple('PlasmaAxis', 'geo mag')
    StrikePoints = namedtuple('StrikePoints', 'ib ob')
    DivertorLines = namedtuple('DivertorLines', 'ib ob ib_long ob_long')
    PlasmaLines = namedtuple('PlasmaLines', 'sep sep_closed div ib2ob')

    if sep_val < 1.0:
        # We have manually overridden the separatrix to exclude the X-points and SOL. We're only able to get a few
        # parameters. We will assume only 1 contour will have this value
        contours = plt.contour(psi_data.R[0], psi_data.Z[:, 0], psi_data.psi_norm, [sep_val]).collections[0].get_paths()
        for contour in contours:
            ls = LineString(contour.vertices)
            if wall.convex_hull.contains(ls):
                ibmp = [ls.bounds[0], mag_axis[1]]
                obmp = [ls.bounds[2], mag_axis[1]]
                geo = [ls.centroid.xy[0][0], ls.centroid.xy[1][0]]
                top = [ls.centroid.xy[0][0], ls.bounds[3]]
                bottom = [ls.centroid.xy[0][0], ls.bounds[1]]
                mag = mag_axis
                pts = Points(ibmp, obmp, top, bottom, xpt,
                             PlasmaAxis(geo, mag),
                             StrikePoints(None, None))
                lines = PlasmaLines(ls, ls, None, None)
                return pts, lines
        raise Exception("Could not find an LCFS with overwritten sep_val")
    #We find the potential sepatartrix lines first
    contours_paths = plt.contour(psi_data.R[0], psi_data.Z[:, 0], psi_data.psi_norm, [sep_val]).collections[0].get_paths()

    if xpt[0] is not None and xpt[1] is not None:
        # We have 2 X-points. We'll basically ignore one, and SOL will be disabled in the rest of GT3. Any neutrals
        # calculations will have to be omitted. We also know that this line (possibly by definition) will contain
        # an x-point and have IB/OB strike points. We'll overwrite contours_paths to say that this is the only
        # contour worth caring about in the rest of the calculation.

        temp_contours = plt.contour(psi_data.R[0], psi_data.Z[:, 0], psi_data.psi_norm, [sep_val]).collections[0].get_paths()
        if len(temp_contours) == 1:
            ls = LineString(temp_contours.vertices)
            if ls.convex_hull.contains(Point(xpt[0])):
                # The upper X-point will define our separatrix
                xpt = xpt[0]
            elif ls.convex_hull.contains(Point(xpt[1])):
                # The lower X-point will define our separatrix
                xpt = xpt[1]
            else:
                raise Exception("There are 2 x-points, but neither are on the separatrix as found in CalcPtsLines.")
        else:
            for path in temp_contours:
                ls = LineString(path.vertices)
                if ls.convex_hull.contains(Point(xpt[0])):
                    # The upper X-point will define our separatrix
                    xpt = xpt[0]
                    contours_paths = path
                    break
                elif ls.convex_hull.contains(Point(xpt[1])):
                    # The lower X-point will define our separatrix
                    xpt = xpt[1]
                    contours_paths = path
                    break
    else:
        # Grab the actual x-point
        if xpt[0] is not None:
            xpt_temp = xpt[0]
        else:
            xpt_temp = xpt[1]

    # create lines for seperatrix and divertor legs of seperatrix

    #Convert back to arrays since I don't want to deal with changing everything to shapely since this already works
    if isinstance(contours_paths, Path):
        contours = [contours_paths.vertices]
    else:
        contours = [a.vertices for a in contours_paths]
    # Replace points in the vic. of the xpt with the xpt
    contour_new = []
    for contour in contours:
        dist = np.sqrt((contour[:, 0] - xpt_temp[0])**2 + (contour[:, 1] - xpt_temp[1])**2)
        # TODO: Put this distance in the input file or make distance selection smarter
        contour[dist < 0.1] = xpt_temp
        contour_new.append(contour[np.where((contour != np.roll(contour, -1, axis=0)).all(axis=1))])

    # if there is only one contour, then it's drawing one continuous line from ib strike point to ob strike point
    # if there are two contours, then it's drawing separate seperatrix and divertor legs

    if len(contour_new) == 1:
        contour = contour_new[0]

        # make sure the line goes from inboard to outboard
        if contour[-1, 0] < contour[0, 0]:
            contour = np.flipud(contour_new)[0]

        xpt_loc = np.where((contour == xpt_temp).all(axis=1))[0]
        ib = np.flipud(contour[:xpt_loc[0]+1])
        lcfs = contour[xpt_loc[0]:xpt_loc[1]]
        ob = contour[xpt_loc[1]:]

    elif len(contour_new) > 1:
        for contour in contour_new:
            # determine if contour is seperatrix, divertor legs, a main ib2ob line, or something else
            contour = np.array(contour)
            if not LineString(contour).intersects(wall) and wall.convex_hull.contains(LineString(contour)):
                lcfs = contour
                continue
            # Meaningless noise lines will not pass near the x-point, so check that first to elliminate them
            if LineString(contour).distance(Point(xpt_temp)) < 0.01:
                # if the largest y value is larger than the y value of the magnetic axis AND the line
                # intersects the wall twice, then it's an ib2ob line and there are more than one psi_norm=1
                # contours because the others are noise. Treat this one the same way we would if there were
                # only one contour.

                # count number of intersections with the wall
                wall_ints = len(LineString(contour).intersection(wall))

                if wall_ints >= 2 and np.amax(contour[:, 1]) > mag_axis[1]:
                    # then we probably have an ib2ob line. Treat the same way as above

                    # make sure the line goes from inboard to outboard
                    if contour[-1, 0] < contour[0, 0]:
                        contour = np.flipud(contour_new)

                    xpt_loc = np.where((contour == xpt_temp).all(axis=1))[0]

                    ib = np.flipud(contour[:xpt_loc[0] + 1])
                    lcfs = contour[xpt_loc[0]:xpt_loc[1]]
                    ob = contour[xpt_loc[1]:]

                # if the largest y value is larger than the y value of the magnetic axis
                elif np.amax(contour[:, 1]) > mag_axis[1]:
                    # we have the main seperatrix
                    lcfs = contour

                else:  # we have the divertor legs
                    # make sure the line goes from inboard to outboard
                    if contour[-1, 0] < contour[0, 0]:
                        contour = np.flipud(contour)

                    # create ib and ob divertor legs, each starting at the x-point
                    xpt_loc = np.where((contour == xpt_temp).all(axis=1))[0][0]
                    ob = contour[xpt_loc:]
                    ib = np.flipud(contour[:xpt_loc+1])

    # create lcfs lines
    lcfs_line = LineString(lcfs)
    lcfs_line_closed = LinearRing(lcfs)

    # create ib and ob linestrings and truncate at the wall
    ib_line = LineString(ib)
    ob_line = LineString(ob)

    ib_strike = ib_line.intersection(wall)
    ob_strike = ob_line.intersection(wall)

    ib_line_cut = cut(ib_line, ib_line.project(ib_strike, normalized=True))[0]
    ob_line_cut = cut(ob_line, ob_line.project(ob_strike, normalized=True))[0]

    entire_sep = np.vstack((np.flipud(ib), lcfs[1:], ob))
    entire_sep_line = LineString(entire_sep)

    # create points object
    obmp_pt = lcfs[np.argmax(lcfs, axis=0)[0]]
    ibmp_pt = lcfs[np.argmin(lcfs, axis=0)[0]]
    top_pt = lcfs[np.argmax(lcfs, axis=0)[1]]
    bot_pt = lcfs[np.argmin(lcfs, axis=0)[1]]
    axis_geo = [(obmp_pt[0] + ibmp_pt[0]) / 2, (obmp_pt[1] + ibmp_pt[1]) / 2]

    pts = Points(
        ibmp_pt,
        obmp_pt,
        top_pt,
        bot_pt,
        xpt,
        PlasmaAxis(
            axis_geo,
            mag_axis),
        StrikePoints(
            ib_strike,
            ob_strike))

    # create lines object
    div_lines = DivertorLines(
        ib_line_cut,
        ob_line_cut,
        ib_line,
        ob_line)

    lines = PlasmaLines(
        lcfs_line,
        lcfs_line_closed,
        div_lines,
        entire_sep_line)

    return pts, lines


#!/usr/bin/python
from collections import namedtuple

def eVConvert(e):
    """
    Given a value in eV, returns an EnergyConversion named tuple containing the value in eV, keV, and J.

    :param e:
    :type e: float
    :return:
    """
    EnergyConversions = namedtuple('EnergyConversion', 'eV keV J')
    return EnergyConversions(e, e * 1E-3, e * 1.0618E-19)


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

def getNum(prompt, valType):
    """

    Function to obtain valid inputs from keyboard

    :param prompt: Text prompt to user
    :param valType: Type of value expected from user
    :return: Your mom
    """

    if valType=='f':
        while True:
            try:
                return float(input(prompt))
            except:
                print("Invalid input \n")
    elif valType=='i':
        while True:
            try:
                x = input(prompt)
                if isinstance(x, (int, long)):
                    return(int(x))
                else:
                    print "Invalid input \n"
                    continue
            except:
                print "Invalid input \n"


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.constants import constants
from math import pi


eps_0 = constants.epsilon_0
e = constants.elementary_charge

def calc_coul_log(z1, z2, T_J, n2):
    """Calculate the coulomb logarithm.

    Reference Equation 1.36 in Dr. Stacey's Fusion Plasma Physics book

    :param z1:
    :param z2:
    :param T_J:
    :param n2:
    :return:
    """

    coul_log = np.log(12 * pi * np.sqrt((eps_0 * T_J)**3 / (n2 * (z2*e)**4 * (z1*e)**2)))
    return coul_log


def calc_coul_log_j_k(z_j, z_k, T_j, n_k):
    # Coulomb logarithm calculation in GTEDGE PARAM subroutine.
    coul_log = np.log(12 * pi * (T_j**1.5) * ((eps_0/e)**1.5) / (np.sqrt(n_k) * (z_k**2.0) * z_j))
    return coul_log


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 20 08:58:46 2018

@author: max
"""

from __future__ import division
import numpy as np
import matplotlib.pyplot as plt
import sys
from GT3 import Core, BeamDeposition
from collections import namedtuple
from math import sqrt
from scipy.interpolate import UnivariateSpline, interp1d
from GT3.RadialTransport.Functions.CorePatch import corePatch
from GT3.RadialTransport.Functions.CalcReturnCur import calc_return_cur
from GT3.RadialTransport.Functions.CalcNu import calc_nu_j_k, calc_nu_drag, calc_nustar
from GT3.RadialTransport.Functions.CalcMbalRHS import calc_mbal_rhs
from GT3.RadialTransport.Functions.CalcT90 import calc_t90
from GT3.RadialTransport.Functions.CalcQ import calc_Qe_diff_method, calc_qie, calc_Qe_int_method, calc_Qi_int_method, \
    calc_Qi_diff_method
from GT3.RadialTransport.Functions.CalcGamma import calc_gamma_diff_method, calc_gamma_int_method
from GT3.RadialTransport.Functions.CalcIntrinRot import calc_intrin_rot
from GT3.RadialTransport.Functions.CalcVTorDPert import calc_vtor_d_pert
from GT3.RadialTransport.Functions.CalcErMomBal import calc_Er_mom_bal
from GT3.RadialTransport.Functions.CalcErIOL import calc_Er_iol
from GT3.RadialTransport.Functions.CalcVpol import calc_vpol
from GT3.RadialTransport.Functions.CalcCXCool import calc_cxcool
import GT3.constants as constants

eps_0 = constants.epsilon_0
e = constants.elementary_charge
m_e = constants.electron_mass
m_d = constants.deuteron_mass
m_c = 12 / constants.N_A / 1E3  # in kg

z_d = 1  # atomic number of deuterium
z_c = 6  # atomic number of carbon
ch_d = e * z_d  # charge of deuterium
ch_c = e * z_c  # charge of carbon

E_phi = 0.04  # toroidal electrostatic potential

MARKERSIZE = constants.MARKERSIZE


def viscCalc(data, core):
    """

    :type core: Core
    :type data: RadialTransport
    """
    fp = core.B_p_fsa / core.B_t_fsa
    ni = core.n_fsa.i
    Ti = core.T_fsa.i.J
    q = core.q[:, 0]
    R0 = core.R0_a
    vth = data.vpol_D[0]
    vtor = data.vtor_D_total
    vpol = data.vpol_D
    eps = core.a / core.R0_a
    nustar = data.nustar[0]
    geom = (eps ** (-3./2.) * nustar) / ((1 + eps ** (-3. / 2.) * nustar) * (1 + nustar))
    vtorS = 0.1
    vpolS = 0.1
    #eta0 = [a * m_d * b * c * core.R0_a * f1 for a, b, c in zip(n.i, vth, core.q[:, 0])]
    eta0 = ni * m_d * vth * q * R0 * geom
    #eta4 = [a * m_d * c * ch_d / (ch_d * abs(b)) for a, b, c in zip(n.i, core.B_t_fsa, T.i.ev)]
    eta4 = ni * m_d * Ti / (ch_d * abs(core.B_t_fsa))
    vrad = data.gamma_diff_D / ni

    # a = vtor    b = fp   c = eta0
    # d = vrad    f = vthet g = eta 4

    #return [a * (b * c * d - .5 * g * (4.0 * a + f)) - .5 * f * (c * d + g * (a + .5 * f)) for a, b, c, d, f, g in
    #        zip(data.vtor_D_total, fp, eta0, vrad, data.vpol_D, eta4)]

    res = vtor * vtorS * (eta0 * fp * vrad - eta4 * (2. * vtor + .5 * vpol))
    res = res - 0.5 * vpol * vpolS * (eta0 * vrad + eta4 * (vtor + .5 * vpol))
    return res / R0

def calc_chi_e(Qe, gamma_diff_D, gamma_C, n, L, T):
    gameltemp = 1.0 * gamma_diff_D + 6.0 * gamma_C   #OHHHH gamma electron!!! HA HA HA HA WE DON'T KNOW WHAT THE FUCK GAMMA_C IS

    return L.T.e * ((Qe / (ch_d * n.e * T.e.ev)) - 2.5 * gameltemp / n.e)

def calc_external_term(M_phi, n_j, ch_j, B_p):
    ext_term = (-M_phi - (n_j * ch_j * E_phi)) / (n_j * ch_j * B_p)
    return ext_term


def calc_poloidal_term(n_j, m_j, ch_j, nu_jk, nu_dj, B_t, B_p, v_pol_j):
    pol_term = (n_j * m_j * (nu_jk + nu_dj) * B_t * v_pol_j) / (n_j * ch_j * (B_p ** 2.0))
    return pol_term


def calc_radial_E_field_term(n_j, m_j, ch_j, nu_jk, nu_dj, Er, B_p):
    Er_term = (n_j * m_j * (nu_jk + nu_dj) * Er) / (n_j * ch_j * (B_p ** 2.0))
    return Er_term


def calc_toroidal_term(n_j, m_j, ch_j, nu_jk, B_p, v_tor_k):
    tor_term = (-n_j * m_j * nu_jk * v_tor_k) / (n_j * ch_j * B_p)
    return tor_term


def calc_pinch_velocity(ext_term, pol_term, Er_term, tor_term):
    vr_pinch = ext_term + pol_term + Er_term + tor_term
    return vr_pinch


class RadialTransport:

    def __init__(self, core, iol, nbi, iolFlag=True, neutFlag=True):
        """

        :type nbi: BeamDeposition
        :type core: Core
        """
        sys.dont_write_bytecode = True

        ##############################################################
        # prep quantities for 1D transport analysis
        ##############################################################

        # prepare beams object
        corePatch(core, neutFlag)  # Patch to update values not brought in via ffiles (ni, zeff)
        #neutPatch(core)
        try:
            dn_dr = core.dn_dr_fsa
        except:
            dn_dr = None

        # prepare core and iol quantities
        r = core.r.T[0]  # TODO: Should this be a flux surface average?
        self.rhor = core.r[:, 0] / core.a
        """The rho vector"""
        self.core = core
        """The utilized GT3 core background plasma"""
        self.nbi = nbi
        """The utilized GT3 NBI module data"""
        self.iol = iol
        """The utilized GT3 IOL module data"""
        self.iolFlag = iolFlag


        self.izn_rate = core.izn_rate_fsa  # TODO: Should this be a flux surface average or a flux surface total?
        self.cool_rate = core.cool_rate_fsa  # TODO: Should this be a flux surface average or a flux surface total?
        n = core.n_fsa

        T = core.T_fsa
        L = core.L_fsa
        B_p = core.B_p_fsa
        B_t = core.B_t_fsa
        Er = core.E_r_fsa  # * 1000.0 # Piper Changes: Convert input Er from kV/m to V/m
        Lp = core.Lp_fsa
        dp_dr = core.dp_dr_fsa

        # prepare iol quantities
        F_orb_d = iol.forb_d_therm_1D
        M_orb_d = iol.morb_d_therm_1D
        E_orb_d = iol.eorb_d_therm_1D

        F_orb_c = iol.forb_c_therm_1D
        M_orb_c = iol.morb_c_therm_1D
        E_orb_c = iol.eorb_c_therm_1D

        F_orb_t = iol.forb_t_therm_1D
        M_orb_t = iol.morb_t_therm_1D
        E_orb_t = iol.eorb_t_therm_1D

        # prepare fast iol quantities
        F_orb_d_nbi = iol.forb_d_nbi_1D
        M_orb_d_nbi = iol.morb_d_nbi_1D
        E_orb_d_nbi = iol.eorb_d_nbi_1D

        ##############################################################
        # particle balance
        ##############################################################

        self.part_src_nbi = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_total.Snbi)(self.rhor))
        self.part_src_nbi_tot = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_total.Snbi)(self.rhor))
        self.part_src_nbi_lost = np.array(interp1d(nbi.beams_space, nbi.combined_beam_src_dens_lost.Snbi, kind="linear")(self.rhor))
        self.part_src_nbi_kept = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_kept.Snbi)(self.rhor))

        # Piper changes: Changed names of particle and heat flux so it's easier to tell what method is used.
        self.gamma_diff_D = calc_gamma_diff_method(r, core.a, self.part_src_nbi, self.part_src_nbi_lost, self.izn_rate, core.dVdrho,
                                                   iol_adjusted=iolFlag,
                                                   F_orb=F_orb_d, neutFlag=neutFlag)  # Differential Cylindrical Method
        self.gamma_int_D = calc_gamma_int_method(r, self.part_src_nbi_tot, self.part_src_nbi_lost, self.izn_rate, iol_adjusted=iolFlag,
                                                 F_orb=F_orb_d, neutFlag=neutFlag)  # Integral Cylindrical Method
        self.gamma_C = np.zeros(self.gamma_int_D.shape)

        # Piper changes: Calculate radial return current (Uses integral cylindrical gamma)
        self.jr_iol = calc_return_cur(r, self.part_src_nbi_lost, self.gamma_int_D, ch_d, iol_adjusted=iolFlag,
                                      F_orb=F_orb_d)
        self.Er_iol, self.iol_term, self.diamag_term, self.diamag_term_orig, self.neut_dens_term = calc_Er_iol(n.i, n.e,
                                                                                                               m_d, n.n,
                                                                                                               B_t,
                                                                                                               Lp.i,
                                                                                                               dp_dr.i,
                                                                                                               e * z_d,
                                                                                                               T.i,
                                                                                                               dn_dr,
                                                                                                               self.izn_rate,
                                                                                                               self.jr_iol)

        ##############################################################
        # momentum balance
        ##############################################################
        self.mom_src_nbi = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_total.Mnbi)(self.rhor))
        self.mom_src_nbi_tot = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_total.Mnbi)(self.rhor))
        self.mom_src_nbi_lost = np.array(interp1d(nbi.beams_space, nbi.combined_beam_src_dens_lost.Mnbi, kind="linear")(self.rhor))
        self.mom_src_nbi_kept = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_kept.Mnbi)(self.rhor))

        # calculate momentum source from anomalous torque
        self.mom_src_anom = np.zeros(r.shape)  # TODO: Anomolous torque

        frac = n.i / (n.i + n.C)
        self.mom_src_tor_D_tot = (1 - frac) * (self.mom_src_nbi + self.mom_src_anom)
        self.mom_src_tor_C_tot = frac * (self.mom_src_nbi + self.mom_src_anom)

        ##############################################################
        # rotation
        ##############################################################

        # calculate carbon toroidal rotation
        self.vtor_C_intrin = calc_intrin_rot(M_orb_c, T.i.J, m_c)
        self.vtor_C_total = core.v_1D.tor.C
        self.vtor_C_fluid = self.vtor_C_total - self.vtor_C_intrin

        # calculate deuterium toroidal rotation
        self.vtor_D_intrin = calc_intrin_rot(M_orb_d, T.i.J, m_d)

        # Piper Changes: Changed core.v_1D.tor.C.any() to core.v_1D.tor.D.any(). Carbon velocity should be a given.
        if not core.v_1D.tor.D.any():  # if array is all zeros, then no input. Use perturbation theory.
            self.vtor_D_total = calc_vtor_d_pert(self.vtor_C_total,
                                                 self.vtor_C_intrin,
                                                 self.vtor_D_intrin,
                                                 self.mom_src_tor_D_tot,
                                                 1,
                                                 n,
                                                 T,
                                                 B_p,
                                                 self.gamma_int_D,
                                                 self.gamma_C)  # Piper Changes: Uses integral cylindrical gamma
            # Piper changes: added a message to let the user know the D velocity was calculated.
            print 'Deuterium toroidal velocity calculated from perturbation theory.'
        else:
            # Piper Changes: For some reason this used to set D velocity to C velocity,
            # which overwrote the input D velocity.
            self.vtor_D_total = core.v_1D.tor.D

        self.vtor_fluid_D = self.vtor_D_total - self.vtor_D_intrin

        # calculate carbon and deuterium poloidal rotation
        try:
            self.vpol_C = core.v_1D.pol.C
            self.vpol_D, self.vpol_D_assum, self.vpol_D_alt = calc_vpol(Er, self.vtor_D_total, Lp, T, n, z_d, B_t, B_p,
                                                                        self.vtor_C_total, self.vpol_C, z_c)
        except:
            self.vpol_D = self.vpol_C / 0.4
            print 'could not calculate deuterium poloidal rotation'
            pass

        # Nick Changes: TEMPORARY - Calculate Er using pressure gradient vs. scale length.
        self.Er_calc_D, self.Er_pres_term_D, self.Er_vxb_term_D = calc_Er_mom_bal(n.i, e * z_d, dp_dr.i, T.i, Lp.i,
                                                                                  self.vtor_D_total, self.vpol_D, B_t,
                                                                                  B_p)
        self.Er_calc_C, self.Er_pres_term_C, self.Er_vxb_term_C = calc_Er_mom_bal(n.C, e * z_c, dp_dr.C, T.C, Lp.C,
                                                                                  self.vtor_C_total, self.vpol_C, B_t,
                                                                                  B_p)

        # calculate nu_drags
        #mbal_rhs_D = calc_mbal_rhs(self.mom_src_tor_D_tot, z_d, n.i, B_p,
        #                           self.gamma_int_D)  # Piper Changes: Uses integral cylindrical gamma
        mbal_rhs_D = calc_mbal_rhs(self.mom_src_tor_D_tot, z_d, n.i, B_p,
                                   self.gamma_diff_D)  # Piper Changes: Uses integral cylindrical gamma
        mbal_rhs_C = calc_mbal_rhs(self.mom_src_tor_C_tot, z_c, n.C, B_p, self.gamma_C)

        nu_c_DC = 1 / calc_t90(m_d, m_c, z_d, z_c, n.C, T.i.J)
        nu_c_CD = 1 / calc_t90(m_c, m_d, z_c, z_d, n.i, T.i.J)

        # Piper changes: added alternate collision frequency calculation for comparison.
        self.nu_c_j_k = calc_nu_j_k(m_d, m_c, z_d, z_c, T.i.ev, n.C)
        self.nu_c_k_j = calc_nu_j_k(m_c, m_d, z_c, z_d, T.C.ev, n.i)
        self.nu_c_j_j = calc_nu_j_k(m_d, m_d, z_d, z_d, T.i.ev, n.i)

        self.nu_drag_D = calc_nu_drag(n.i, m_d, self.vtor_D_total, self.vtor_C_total, mbal_rhs_D, nu_c_DC)
        self.nu_drag_C = calc_nu_drag(n.i, m_d, self.vtor_D_total, self.vtor_C_total, mbal_rhs_C, nu_c_CD)
        self.nustar = calc_nustar(self.nu_c_j_j, core.q[:, 0], core.R0_a, self.vpol_D)

        ##############################################################
        # Pinch Velocity
        ##############################################################

        # Piper Changes: Added pinch velocity section and calculations.


        self.vrpinch_ext_term = calc_external_term(self.mom_src_tor_D_tot, n.i, ch_d, B_p)
        self.vrpinch_poloidal_term = calc_poloidal_term(n.i, m_d, ch_d, nu_c_DC, self.nu_drag_D, B_t, B_p, self.vpol_D)
        self.vrpinch_Er_term = calc_radial_E_field_term(n.i, m_d, ch_d, nu_c_DC, self.nu_drag_D, Er, B_p)
        self.vrpinch_toroidal_term = calc_toroidal_term(n.i, m_d, ch_d, nu_c_DC, B_p, self.vtor_C_total)
        self.vrpinch = calc_pinch_velocity(self.vrpinch_ext_term, self.vrpinch_poloidal_term, self.vrpinch_Er_term,
                                           self.vrpinch_toroidal_term)

        ##############################################################
        # energy balance
        ##############################################################

        # I don't honestly understand why anything but densities are used anywhere. Changed to just use densities.

        part_src_nbi = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_total.Snbi)(self.rhor))
        part_src_nbi_tot = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_total.Snbi)(self.rhor))
        part_src_nbi_lost = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_lost.Snbi)(self.rhor))
        part_src_nbi_kept = np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_kept.Snbi)(self.rhor))

        self.en_src_nbi_i = 0.5 * np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_total.Qnbi)(self.rhor))
        self.en_src_nbi_i_tot = 0.5 * np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_total.Qnbi)(self.rhor))
        self.en_src_nbi_i_lost = 0.5 * np.array(interp1d(nbi.beams_space, nbi.combined_beam_src_dens_lost.Qnbi, kind="linear")(self.rhor))
        self.en_src_nbi_i_kept = 0.5 * np.array(UnivariateSpline(nbi.beams_space, nbi.combined_beam_src_dens_kept.Qnbi)(self.rhor))

        ################################################################################################################
        #
        #   NBI energy split - Currently 50:50 split ions and electrons
        #
        #   TODO: Implement accurate split
        #
        ################################################################################################################

        self.en_src_nbi_e = self.en_src_nbi_i_kept
        self.cxcool = calc_cxcool(core, n, T)
        self.qie = calc_qie(n, T, ion_species='D')

        # calculate radial heat flux. Piper changes: Separated heat flux equations into differential and integral cylindrical methods.
        self.Qi_diff = calc_Qi_diff_method(r, core.a, self.cxcool, self.qie, self.en_src_nbi_i_tot,
                                           self.en_src_nbi_i_lost, core.dVdrho, iol_adjusted=iolFlag,
                                           E_orb=E_orb_d)  # previously called qheat. Differential Method.

        self.Qi_int = calc_Qi_int_method(r, n, T, self.qie, self.en_src_nbi_i_kept, self.cool_rate, iol_adjusted=iolFlag,
                                                   E_orb=E_orb_d)  # Integral method.
        self.Qe_diff = calc_Qe_diff_method(r, core.a, self.en_src_nbi_e, self.cool_rate, calc_qie(n, T))  # Differential Method.

        self.Qe_int = calc_Qe_int_method(r, n, T, self.en_src_nbi_e, self.cool_rate)  # Integral method.

        self.conv15 = UnivariateSpline(core.r[:, 0], 3. * .5 * ch_d * self.gamma_diff_D * T.i.ev, k=3, s=0)(core.r[:, 0])
        self.conv25 = UnivariateSpline(core.r[:, 0], 5. * .5 * ch_d * self.gamma_diff_D * T.i.ev, k=3, s=0)(core.r[:, 0])

        self.heatvisc = viscCalc(self, core)

        self.heatin = UnivariateSpline(core.r[:, 0],
                                       .5 * self.gamma_diff_D * m_d * (self.vtor_D_total ** 2 + self.vpol_D ** 2), k=3,
                                       s=0)(core.r[:, 0])  # TODO: Provide logic that uses vtor_D_intrin/fluid depending on IOL Switch, currently too small to matter

        self.chi = namedtuple('chi', 'i e')(
            namedtuple('i', 'chi1 chi2 chi3 chi4')(
                (self.Qi_diff) * L.T.i / (n.i * T.i.ev * ch_d),
                (self.Qi_diff - self.conv25) * L.T.i / (n.i * T.i.ev * ch_d),
                (self.Qi_diff - self.conv25 - self.heatin) * L.T.i / (n.i * T.i.ev * ch_d),
                (self.Qi_diff - self.conv25 - self.heatin - self.heatvisc) * L.T.i / (n.i * T.i.ev * ch_d)
            ), calc_chi_e(self.Qe_diff, self.gamma_diff_D, self.gamma_C, n, L, T)
        )

        self.D_i = m_d * core.T_fsa.i.J * (self.nu_c_j_k * (1. - ch_d / ch_c) + self.nu_drag_D) / ((ch_d * core.B_p_fsa)**2)

    def plot_chi_terms(self, edge=True):
        fig = self._plot_base(self.conv25, title="", yLabel="q[W/m^2]", edge=edge)
        fig.scatter(self.rhor, self.heatin, color="blue", s=MARKERSIZE)
        fig.scatter(self.rhor, self.heatvisc, color="purple", s=MARKERSIZE)
        fig.scatter(self.rhor, self.Qi_diff, color="black", s=MARKERSIZE)
        #fig.legend([r"$q^{conv}$", r"$q^{heatin}$", r"$q^{tot}$"])
        fig.legend([r"$q^{conv}$", r"$q^{heatin}$", r"$q^{visc}$", r"$q^{tot}$"])
        return fig

    def plot_gamma(self, edge=True):
        fig = self._plot_base(self.gamma_int_D, title="Radial particle flux", yLabel=r"$\Gamma[W/m^3]$", edge=edge)
        fig.scatter(self.rhor, self.gamma_diff_D, color="blue", s=MARKERSIZE)
        fig.legend([r"$\Gamma_{r, int}$", r"$\Gamma_{r, diff}$"])
        return fig

    def plot_gamma_diff(self, edge=True):
        fig = self._plot_base(self.gamma_diff_D, title="Radial particle flux", yLabel=r"$\Gamma[W/m^3]$", edge=edge)
        return fig

    def plot_gamma_int(self, edge=True):
        fig = self._plot_base(self.gamma_int_D, title="Radial particle flux", yLabel=r"$\Gamma[W/m^3]$", edge=edge)
        return fig

    def plot_gamma_diff_calc(self):
        calc_gamma_diff_method(self.rhor * self.core.a,
                               self.core.a,
                               self.part_src_nbi,
                               self.part_src_nbi_lost,
                               self.izn_rate,
                               self.core.dVdrho,
                               iol_adjusted=self.iolFlag,
                               F_orb=self.iol.forb_d_therm_1D,
                               verbose=True)

    def plot_Q_i(self, edge=True):
        fig = self._plot_base(self.Qi_int, yLabel=r'$Q_i [W/m^2]$', title="Ion heat flux", edge=edge)
        fig.scatter(self.rhor, self.Qi_diff, color="black")
        fig.legend([r"$Q^{int}_i$", r"$Q^{diff}_i$"])
        return fig

    def plot_Q_e_int(self, edge=True):
        return self._plot_base(self.Qe_int, yLabel=r'$Q_e^{int} [W/m^2]$', title="Electron heat flux", edge=edge)

    def plot_q_diff_calc(self):
        calc_Qi_diff_method(self.rhor * self.core.a,
                            self.core.a,
                            self.cxcool,
                            self.qie,
                            self.en_src_nbi_i_tot,
                            self.en_src_nbi_i_lost,
                            self.core.dVdrho,
                            iol_adjusted=self.iolFlag,
                            E_orb=self.iol.eorb_d_therm_1D,
                            verbose=True)

    def _plot_base(self, val, xLabel=r'$\rho$', yLabel="Value", title="Title", color='red', edge=False):

        plot = plt.figure()
        fig = plot.add_subplot(111)
        fig.set_xlabel(xLabel, fontsize=30)
        fig.set_ylabel(yLabel, fontsize=30)
        plt.xticks(fontsize=20)
        plt.yticks(fontsize=20)
        fig.set_title(title)
        if edge:
            fig.set_xlim(0.85, 1.0)
        fig.scatter(self.rhor, val, color=color, s=MARKERSIZE)
        plt.show()
        return fig

    def plot_ni(self, edge=True):
        return self._plot_base(self.core.n_fsa.i, yLabel=r'$n_i[\#/m^3]$', title="Ion Density", edge=edge)

    def plot_ne(self, edge=True):
        return self._plot_base(self.core.n_fsa.e, yLabel=r'$n_e[\#/m^3]$', title="Electron Density", edge=edge)

    def plot_n(self, edge=True):
        fig = self._plot_base(self.core.n_fsa.e, yLabel=r'$n_e[\#/m^3]$', title="Plasma Densities", edge=edge)
        fig.scatter(self.rhor, self.core.n_fsa.i, color="blue", s=MARKERSIZE)
        return fig

    def plot_nn_therm(self, edge=True):
        return self._plot_base(self.core.n_fsa.n.t, yLabel=r'$n_{n,therm}[\#/m^3]$', title="Thermal Neutral Density", edge=edge)

    def plot_nn_cold(self, edge=True):
        return self._plot_base(self.core.n_fsa.n.s, yLabel=r'$n_{n,slow}[\#/m^3]$', title="Cold Neutral Density", edge=edge)

    def plot_nn_total(self, edge=True):
        return self._plot_base(self.core.n_fsa.n.tot, yLabel=r'$n_{n,total}[\#/m^3]$', title="Total Neutral Density", edge=edge)

    def plot_Ti(self, edge=True):
        return self._plot_base(self.core.T_fsa.i.kev, yLabel=r'$T_i[keV]$', title="Ion Temperature", edge=edge)

    def plot_Te(self, edge=True):
        return self._plot_base(self.core.T_fsa.e.kev, yLabel=r'$T_e[keV]$', title="Electron Temperature", edge=edge)

    def plot_T(self, edge=True):
        fig = self._plot_base(self.core.T_fsa.e.kev, yLabel=r'$T[keV]$', title="Plasma Temperature", edge=edge)
        fig.scatter(self.rhor, self.core.T_fsa.i.kev,  color="blue", s=MARKERSIZE)
        plt.show()
        return fig

    def plot_Er(self, edge=True):
        return self._plot_base(self.Er_calc_C, yLabel=r'$E_r[V/m]$', title="Radial Electric Field", edge=edge)

    def plot_S_sources(self, edge=True, logPlot=True):
        fig = self._plot_base(self.part_src_nbi, yLabel=r'$S_r[#/m^3s]$', title="Radial sources", edge=edge)
        fig.scatter(self.rhor, self.izn_rate, color="green", s=MARKERSIZE)
        if logPlot:
            fig.set_yscale("log")

        fig.legend([r"$S_{nbi}$", r"$S_{izn}$"])
        plt.show()
        return fig

    def plot_Q_sources(self, edge=True, logPlot=False):
        fig = self._plot_base(self.en_src_nbi_i, yLabel=r'$Q_r[W/m^3]$', title="Radial sources", edge=edge)
        fig.scatter(self.rhor, self.cool_rate, color="green", s=MARKERSIZE)
        fig.scatter(self.rhor, self.qie, color="black", s=MARKERSIZE)
        if logPlot:
            fig.set_yscale("log")
        plt.show()
        fig.legend(["Q_{nbi}", "Q_{cxcool}", "Q_{ie}"])
        return fig

    def plot_Chi_i_comp(self, edge=True):
        fig = self._plot_base(self.chi.i.chi1, yLabel=r'$\chi_{r,i}$', title="", edge=edge)
        fig.scatter(self.rhor, self.chi.i.chi2, color="blue", s=MARKERSIZE)
        fig.scatter(self.rhor, self.chi.i.chi3, color="green", s=MARKERSIZE)
        fig.scatter(self.rhor, self.chi.i.chi4, color="purple", s=MARKERSIZE)
        fig.legend([r"$q^{cond} = q^{tot}$",
                    r"$q^{cond} = q^{tot}-q^{conv}$",
                    r"$q^{cond} = q^{tot}-q^{conv}-q^{heatin}$",
                    r"$q^{cond} = q^{tot}-q^{conv}-q^{heatin}-q^{visc}$"],  prop={'size': 20})
        return fig

    def plot_D(self, edge=True):
        fig = self._plot_base(self.D_i, yLabel=r'$D_{r, i} [m^2/s]$', title='', edge=edge)
        return fig


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# isclose is included in python3.5+, so you can delete this if the code ever gets ported into python3.5+
def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


#!/usr/bin/python

try:
    import neutpy
    print "importing neutpy"
except:
    print """
    #############################################
    NeutPy not found. You will be unable to run neutrals calculations without NeutPy.
    
    To install, use 
    
    $ pip install neutpy
    #############################################"""

from gt3 import *



#!/usr/bin/env python2
# -*- coding: utf-8 -*-


def calc_Er_mom_bal(n, e_z, dp_dr, T, Lp, v_tor, v_pol, B_tor, B_pol):
    pres_term = -1.0 / (n * e_z) * dp_dr
    # pres_term_simp = -1.0 * (Lp * T.J) / e_z # alternative method using grad scale length.
    vxb_term = -1.0 * (v_pol * B_tor - v_tor * B_pol)
    E_r = pres_term + vxb_term
    # E_r_simp = pres_term_simp + vxb_term
    return E_r, pres_term, vxb_term


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Sat May 19 14:27:42 2018

@author: max
"""
from __future__ import division
from neutpy import neutrals
from collections import namedtuple
import json
import os.path
from GT3.Core.Processors import NumpyEncoder



class Neutrals:

    def __init__(self, inp, core, cpus=False):


        if abs(1.0 - core.sep_val) > .0001:
            print "The separatrix value has been overwritten. Cannot run Neutrals calculation"
            return
        # Try to read in specified neutrals data file. If it's not there, then prepare inputs for and run neutpy
        self.NeutralDataNT = namedtuple('NeutralsData', 'R Z n_n_slow n_n_thermal izn_rate_slow izn_rate_thermal')
        self.inp = inp
        self.core = core
        try:
            with open(inp.neutfile_loc, "r") as f:
                ntrl_data = json.load(f)
                self.data = self.NeutralDataNT(ntrl_data['R'],
                                        ntrl_data['Z'],
                                        ntrl_data['nn_s_raw'],
                                        ntrl_data['nn_t_raw'],
                                        ntrl_data['izn_rate_slow'],
                                        ntrl_data['izn_rate_thermal'])
        except:
            # Run NeutPy
            print "Neutrals data not found. Running NeutPy"
            if not self._check_conf("neutpy.conf"):
                raise IOError("No NeutPy main configuration file found.")
            self.npi = neutrals()
            if cpus:
                self.npi.set_cpu_cores(cpus)

            self.npi.from_gt3(core, inp)
            self.data = self.NeutralDataNT(self.npi.midpts[:, 0],
                                    self.npi.midpts[:, 1],
                                    self.npi.nn_s_raw,
                                    self.npi.nn_t_raw,
                                    self.npi.iznrate_s_raw,
                                    self.npi.iznrate_t_raw)
            # Save data
            self._save_data()
        try:
            self.core.update_ntrl_data(self.data)
        except:
            print 'unable to update values in core instance.'
            pass

    def reRun(self, cpus=False):
        print ("Manually re-running NeutPy")
        if not self._check_conf("neutpy.conf"):
            raise IOError("No NeutPy main configuration file found.")
        self.npi = neutrals()
        if cpus:
            self.npi.set_cpu_cores(cpus)
        self.npi.from_gt3(self.core, self.inp)
        self.data = self.NeutralDataNT(self.npi.midpts[:, 0],
                                self.npi.midpts[:, 1],
                                self.npi.nn_s_raw,
                                self.npi.nn_t_raw,
                                self.npi.iznrate_s_raw,
                                self.npi.iznrate_t_raw)
        self._save_data()
        try:
            self.core.update_ntrl_data(self.data)
        except:
            print 'unable to update values in core instance.'
            pass

    def _save_data(self):
        try:
            out_dict = {}
            with open(self.inp.neutfile_loc, "w") as f:
                out_dict['R'] = self.data.R
                out_dict['Z'] = self.data.Z
                out_dict['nn_s_raw'] = self.data.n_n_slow
                out_dict['nn_t_raw'] = self.data.n_n_thermal
                out_dict['izn_rate_slow'] = self.data.izn_rate_slow
                out_dict['izn_rate_thermal'] = self.data.izn_rate_thermal
                json.dump(out_dict, f, indent=4, cls=NumpyEncoder)
        except Exception as e:
            print "Unable to save NeutPy data to file: %s" % str(e)

    def _check_conf(self, f):
        """
        Checks to see if a neutpy.conf file exists
        :return: A boolean of whether the file exists.
        :rtype: bool
        """
        return os.path.isfile(f)

